import os
import shutil

from populus import ASSETS_DIR

from populus.config.helpers import (
    check_if_json_config_file_exists,
)

from populus.utils.filesystem import (
    ensure_path_exists,
)

from populus.project import (
    Project,
)

GREETER_SOURCE_PATH = os.path.join(ASSETS_DIR, 'Greeter.sol')
GREETER_TEST_PATH = os.path.join(ASSETS_DIR, 'test_greeter.py')


def init_project(project_dir, logger):

    if project_dir is None:
        project_dir = os.getcwd()
    else:
        project_dir = os.path.abspath(project_dir)

    has_json_config = check_if_json_config_file_exists(project_dir)

    if has_json_config:
        logger.info(
            "Found existing `project.json` file.  Not writing default config."
        )

    project = Project(project_dir, create_config_file=True)
    logger.info(
        "Wrote default populus configuration to `./{0}`.".format(
            os.path.relpath(project.config_file_path),
        )
    )

    for source_dir in project.contracts_source_dirs:
        if ensure_path_exists(source_dir):
            logger.info(
                "Created Directory: ./{0}".format(
                    os.path.relpath(source_dir)
                )
            )

    example_contract_path = os.path.join(project.contracts_source_dirs[0], 'Greeter.sol')
    if not os.path.exists(example_contract_path):
        shutil.copy(GREETER_SOURCE_PATH, example_contract_path)
        logger.info("Created Example Contract: ./{0}".format(
            os.path.relpath(example_contract_path)
        ))

    tests_dir = os.path.join(project.project_dir, 'tests')
    if ensure_path_exists(tests_dir):
        logger.info("Created Directory: ./{0}".format(os.path.relpath(tests_dir)))

    example_tests_path = os.path.join(tests_dir, 'test_greeter.py')
    if not os.path.exists(example_tests_path):
        shutil.copy(GREETER_TEST_PATH, example_tests_path)
        logger.info("Created Example Tests: ./{0}".format(
            os.path.relpath(example_tests_path)
        ))

    return project



import pprint

from cytoolz import (
    assoc,
    concatv,
    partial,
    pipe,
)

from semantic_version import (
    Spec,
)

from eth_utils import (
    add_0x_prefix,
    to_dict,
    to_tuple,
)

from solc import (
    get_solc_version,
    compile_standard,
)
from solc.exceptions import (
    ContractsNotFound,
)

from populus.utils.compile import (
    load_json_if_string,
    normalize_contract_metadata,
)
from populus.utils.linking import (
    normalize_standard_json_link_references,
)
from populus.utils.mappings import (
    has_nested_key,
    get_nested_key,
    set_nested_key,
)

from .base import (
    BaseCompilerBackend,
)


@to_dict
def build_standard_input_sources(source_file_paths):
    for file_path in source_file_paths:
        with open(file_path) as source_file:
            yield file_path, {'content': source_file.read()}


@to_dict
def normalize_standard_json_contract_data(contract_data):
    if 'metadata' in contract_data:
        yield 'metadata', normalize_contract_metadata(contract_data['metadata'])
    if 'evm' in contract_data:
        evm_data = contract_data['evm']
        if 'bytecode' in evm_data:
            yield 'bytecode', add_0x_prefix(evm_data['bytecode'].get('object', ''))
            if 'linkReferences' in evm_data['bytecode']:
                yield 'linkrefs', normalize_standard_json_link_references(
                    evm_data['bytecode']['linkReferences'],
                )
        if 'deployedBytecode' in evm_data:
            yield 'bytecode_runtime', add_0x_prefix(evm_data['deployedBytecode'].get('object', ''))
            if 'linkReferences' in evm_data['deployedBytecode']:
                yield 'linkrefs_runtime', normalize_standard_json_link_references(
                    evm_data['deployedBytecode']['linkReferences'],
                )
    if 'abi' in contract_data:
        yield 'abi', load_json_if_string(contract_data['abi'])
    if 'userdoc' in contract_data:
        yield 'userdoc', load_json_if_string(contract_data['userdoc'])
    if 'devdoc' in contract_data:
        yield 'devdoc', load_json_if_string(contract_data['devdoc'])


@to_tuple
def normalize_compilation_result(compilation_result):
    """
    Take the result from the --standard-json compilation and flatten it into an
    iterable of contract data dictionaries.
    """
    for source_path, file_contracts in compilation_result['contracts'].items():
        for contract_name, raw_contract_data in file_contracts.items():
            contract_data = normalize_standard_json_contract_data(raw_contract_data)
            yield pipe(
                contract_data,
                partial(assoc, key='source_path', value=source_path),
                partial(assoc, key='name', value=contract_name),
            )


REQUIRED_OUTPUT_SELECTION = [
    'abi',
    'metadata',
    'evm.bytecode',
    'evm.bytecode.object',
    'evm.bytecode.linkReferences',
    'evm.deployedBytecode',
    'evm.deployedBytecode.object',
    'evm.deployedBytecode.linkReferences',
]
OUTPUT_SELECTION_KEY = 'settings.outputSelection.*.*'


class SolcStandardJSONBackend(BaseCompilerBackend):
    project_source_glob = ('*.sol', )
    test_source_glob = ('Test*.sol', )

    def __init__(self, *args, **kwargs):
        if get_solc_version() not in Spec('>=0.4.11'):
            raise OSError(
                "The 'SolcStandardJSONBackend can only be used with solc "
                "versions >=0.4.11.  The SolcCombinedJSONBackend should be used "
                "for all versions <=0.4.8"
            )
        super(SolcStandardJSONBackend, self).__init__(*args, **kwargs)

    def get_compiled_contracts(self, source_file_paths, import_remappings):
        self.logger.debug("Import remappings: %s", import_remappings)
        self.logger.debug("Compiler Settings PRE: %s", pprint.pformat(self.compiler_settings))

        # DEBUG
        self.compiler_settings['output_values'] = []
        self.logger.debug("Compiler Settings POST: %s", pprint.pformat(self.compiler_settings))

        if 'remappings' in self.compiler_settings and import_remappings is not None:
            self.logger.warn("Import remappings setting will by overridden by backend settings")

        sources = build_standard_input_sources(source_file_paths)

        std_input = {
            'language': 'Solidity',
            'sources': sources,
            'settings': {
                'remappings': import_remappings,
                'outputSelection': {
                    '*': {
                        '*': REQUIRED_OUTPUT_SELECTION
                    }
                }
            }
        }

        # solc command line options as passed to solc_wrapper()
        # https://github.com/ethereum/py-solc/blob/3a6de359dc31375df46418e6ffd7f45ab9567287/solc/wrapper.py#L20
        command_line_options = self.compiler_settings.get("command_line_options", {})

        # Get Solidity Input Description settings section
        # http://solidity.readthedocs.io/en/develop/using-the-compiler.html#input-description
        std_input_settings = self.compiler_settings.get("stdin", {})
        std_input['settings'].update(std_input_settings)

        # Make sure the output selection has all of the required output values.
        if has_nested_key(std_input, OUTPUT_SELECTION_KEY):
            current_selection = get_nested_key(std_input, OUTPUT_SELECTION_KEY)
            output_selection = list(set(concatv(current_selection, REQUIRED_OUTPUT_SELECTION)))
        else:
            output_selection = REQUIRED_OUTPUT_SELECTION

        set_nested_key(std_input, OUTPUT_SELECTION_KEY, output_selection)

        self.logger.debug("std_input sections: %s", std_input.keys())
        self.logger.debug("Input Description JSON settings are: %s", std_input["settings"])
        self.logger.debug("Command line options are: %s", command_line_options)
        try:
            compilation_result = compile_standard(std_input, **command_line_options)
        except ContractsNotFound:
            return {}

        compiled_contracts = normalize_compilation_result(compilation_result)

        return compiled_contracts



import itertools

import toposort

from populus.utils.contracts import (
    compute_direct_dependency_graph,
    compute_recursive_contract_dependencies,
)


def compute_deploy_order(dependency_graph):
    """
    Given a dictionary that maps contract to their dependencies,
    determine the overall dependency ordering for that set of contracts.
    """
    return toposort.toposort_flatten(dict(dependency_graph))


def get_deploy_order(contracts_to_deploy, compiled_contracts):
    # Extract and dependencies that exist due to library linking.
    dependency_graph = compute_direct_dependency_graph(compiled_contracts.values())
    global_deploy_order = compute_deploy_order(dependency_graph)

    # Compute the full set of dependencies needed to deploy the desired
    # contracts.
    all_deploy_dependencies = set(itertools.chain.from_iterable(
        compute_recursive_contract_dependencies(contract_name, dependency_graph)
        for contract_name in contracts_to_deploy
    ))
    all_contracts_to_deploy = all_deploy_dependencies.union(contracts_to_deploy)

    # Now compute the order that the contracts should be deployed based on
    # their dependencies.
    deploy_order = tuple(
        contract_name
        for contract_name
        in global_deploy_order
        if contract_name in all_contracts_to_deploy
    )
    return deploy_order



from __future__ import absolute_import

import logging
import time

import click

from populus.utils.chains import (
    is_synced,
)
from populus.utils.cli import (
    select_chain,
    deploy_contract_and_verify,
    select_project_contract,
)
from populus.utils.deploy import (
    get_deploy_order,
)


def echo_post_deploy_message(web3, deployed_contracts):
    logger = logging.getLogger('populus.cli.deploy.echo_post_deploy_message')

    message = (
        "========== Deploy Completed ==========\n"
        "Deployed {n} contracts:"
    ).format(
        n=len(deployed_contracts),
    )
    logger.info(message)
    for contract_name, deployed_contract in deployed_contracts:
        deploy_receipt = web3.eth.getTransactionReceipt(deployed_contract.deploy_txn_hash)
        gas_used = deploy_receipt['gasUsed']
        deploy_txn = web3.eth.getTransaction(deploy_receipt['transactionHash'])
        gas_provided = deploy_txn['gas']
        logger.info("- {0} ({1}) gas: {2} / {3}".format(
            contract_name,
            deployed_contract.address,
            gas_used,
            gas_provided,
        ))


def deploy(project, logger, chain_name, wait_for_sync, contracts_to_deploy):
    """
    Deploys the specified contracts to a chain.
    """

    # Determine which chain should be used.
    if not chain_name:
        chain_name = select_chain(project)

    compiled_contracts = project.compiled_contract_data

    if contracts_to_deploy:
        # validate that we *know* about all of the contracts
        unknown_contracts = set(contracts_to_deploy).difference(
            compiled_contracts.keys()
        )
        if unknown_contracts:
            unknown_contracts_message = (
                "Some contracts specified for deploy were not found in the "
                "compiled project contracts.  These contracts could not be found "
                "'{0}'.  Searched these known contracts '{1}'".format(
                    ', '.join(sorted(unknown_contracts)),
                    ', '.join(compiled_contracts.keys()),
                )
            )
            raise click.ClickException(unknown_contracts_message)
    else:
        # prompt the user to select the desired contracts they want to deploy.
        # Potentially display the currently deployed status.
        contracts_to_deploy = [select_project_contract(project)]

    with project.get_chain(chain_name) as chain:
        provider = chain.provider
        registrar = chain.registrar

        # wait for the chain to start syncing.
        if wait_for_sync:
            logger.info("Waiting for chain to start syncing....")
            while chain.wait.for_syncing() and is_synced(chain.web3):
                time.sleep(1)
            logger.info("Chain sync complete")

        # Get the deploy order.
        deploy_order = get_deploy_order(
            contracts_to_deploy,
            compiled_contracts,
        )

        # Display Start Message Info.
        starting_msg = (
            "Beginning contract deployment.  Deploying {0} total contracts ({1} "
            "Specified, {2} because of library dependencies)."
            "\n\n" +
            (" > ".join(deploy_order))
        ).format(
            len(deploy_order),
            len(contracts_to_deploy),
            len(deploy_order) - len(contracts_to_deploy),
        )
        logger.info(starting_msg)

        for contract_name in deploy_order:
            if not provider.are_contract_dependencies_available(contract_name):
                raise ValueError(
                    "Something is wrong with the deploy order.  Some "
                    "dependencies for {0} are not "
                    "available.".format(contract_name)
                )

            # Check if we already have an existing deployed version of that
            # contract (via the registry).  For each of these, prompt the user
            # if they would like to use the existing version.
            if provider.is_contract_available(contract_name):
                # TODO: this block should be a standalone cli util.
                # TODO: this block needs to use the `Provider` API
                existing_contract_instance = provider.get_contract(contract_name)
                found_existing_contract_prompt = (
                    "Found existing version of {name} in registrar. "
                    "Would you like to use the previously deployed "
                    "contract @ {address}?".format(
                        name=contract_name,
                        address=existing_contract_instance.address,
                    )
                )
                if click.prompt(found_existing_contract_prompt, default=True):
                    continue

            # We don't have an existing version of this contract available so
            # deploy it.
            contract_instance = deploy_contract_and_verify(
                chain,
                contract_name=contract_name,
            )

            # Store the contract address for linking of subsequent deployed contracts.
            registrar.set_contract_address(contract_name, contract_instance.address)

        # TODO: fix this message.
        success_msg = (
            "Deployment Successful."
        )
        logger.info(success_msg)



from .solc_combined_json import (  # noqa: F401
    SolcCombinedJSONBackend,
)
from.solc_standard_json import (  # noqa: F401
    SolcStandardJSONBackend,
)
from .solc_auto import (  # noqa: F401
    SolcAutoBackend,
)
from .viper import (  # noqa: F401
    ViperBackend,
)
from .lll import (  # noqa: F401
    LLLBackend,
)



from __future__ import absolute_import

from pylru import lrucache

from eth_utils import (
    to_ordered_dict,
)

from populus.config.backend import (
    ContractBackendConfig,
)

from populus.contracts.provider import (
    Provider,
)
from populus.contracts.registrar import (
    Registrar,
)
from populus.wait import (
    Wait,
)

from populus.config.helpers import (
    sort_prioritized_configs,
)
from populus.utils.functional import (
    cached_property,
)
from populus.utils.module_loading import (
    import_string,
)


class BaseChain(object):
    """
    Base class for how populus interacts with the blockchain.

    :param project: Instance of :class:`populus.project.Project`
    """
    project = None
    chain_name = None
    config = None
    _factory_cache = None

    def __init__(self, project, chain_name, chain_config):
        self.project = project
        self.chain_name = chain_name
        self.config = chain_config
        self._factory_cache = lrucache(128)
        self.initialize_chain()

    def initialize_chain(self):
        """
        Hook for initialization that is called during class instantiation.
        """
        pass

    #
    # Running the chain
    #
    _running = None

    def __enter__(self):
        self._running = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._running:
            raise ValueError("The Chain is not running")
        self._running = False

    #
    # Chain Interaction API
    #
    def get_web3_config(self):
        """
        Return the config object for the web3 instance used by this chain.
        """
        web3_config = self.config.get_web3_config()
        return web3_config

    @property
    def web3_config(self):
        return self.get_web3_config()

    @cached_property
    def web3(self):
        if not self._running:
            raise ValueError("Chain must be running prior to accessing web3")
        return self.web3_config.get_web3()

    @property
    def wait(self):
        return Wait(self.web3)

    #
    # +--------------+
    # | Contract API |
    # +--------------+
    #
    @property
    @to_ordered_dict
    def contract_backend_configs(self):
        backend_configs = self.config.get_config('contracts.backends')
        sorted_backend_configs = sort_prioritized_configs(
            backend_configs,
            self.project.config
        )
        for backend_name, base_backend_config in sorted_backend_configs.items():
            yield backend_name, ContractBackendConfig(base_backend_config)

    @cached_property
    @to_ordered_dict
    def contract_backends(self):
        for backend_name, backend_config in self.contract_backend_configs.items():
            ProviderBackendClass = import_string(backend_config['class'])
            yield (
                backend_name,
                ProviderBackendClass(self, backend_config.get_config('settings')),
            )

    #
    # Provider
    #
    @property
    @to_ordered_dict
    def provider_backends(self):
        for backend_name, backend in self.contract_backends.items():
            if backend.is_provider:
                yield backend_name, backend

    @property
    def provider(self):
        if not self.provider_backends:
            raise ValueError(
                "Must have at least one provider backend "
                "configured\n{0}".format(self.contract_backend_configs)
            )
        return Provider(self, self.provider_backends)

    #
    # Registrar
    #
    @cached_property
    @to_ordered_dict
    def registrar_backends(self):
        for backend_name, backend in self.contract_backends.items():
            if backend.is_registrar:
                yield backend_name, backend

    @property
    def registrar(self):
        if not self.registrar_backends:
            raise ValueError(
                "Must have at least one registrar backend "
                "configured\n{0}".format(self.contract_backend_configs)
            )
        return Registrar(self, self.registrar_backends)



import functools
import random
import time

from web3.providers.tester import (
    EthereumTesterProvider,
    TestRPCProvider,
)


class Timeout(Exception):
    """
    A limited subset of the `gevent.Timeout` context manager.
    """
    seconds = None
    exception = None
    begun_at = None
    is_running = None

    def __init__(self, seconds=None, exception=None, *args, **kwargs):
        self.seconds = seconds
        self.exception = exception

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def __str__(self):
        if self.seconds is None:
            return ''
        return "{0} seconds".format(self.seconds)

    @property
    def expire_at(self):
        if self.seconds is None:
            raise ValueError("Timeouts with `seconds == None` do not have an expiration time")
        elif self.begun_at is None:
            raise ValueError("Timeout has not been started")
        return self.begun_at + self.seconds

    def start(self):
        if self.is_running is not None:
            raise ValueError("Timeout has already been started")
        self.begun_at = time.time()
        self.is_running = True

    def check(self):
        if self.is_running is None:
            raise ValueError("Timeout has not been started")
        elif self.is_running is False:
            raise ValueError("Timeout has already been cancelled")
        elif self.seconds is None:
            return
        elif time.time() > self.expire_at:
            self.is_running = False
            if isinstance(self.exception, type):
                raise self.exception(str(self))
            elif isinstance(self.exception, Exception):
                raise self.exception
            else:
                raise self

    def cancel(self):
        self.is_running = False

    def sleep(self, seconds):
        time.sleep(seconds)
        self.check()


def poll_until(poll_fn, success_fn, timeout, poll_interval_fn):
    with Timeout(timeout) as _timeout:
        while True:
            value = poll_fn()

            if success_fn(value):
                return value

            _timeout.sleep(poll_interval_fn())


def is_tester_web3(web3):
    return isinstance(web3.providers[0], (TestRPCProvider, EthereumTesterProvider))


def wait_for_transaction_receipt(web3, txn_hash, timeout=120, poll_interval=None):
    return poll_until(
        poll_fn=functools.partial(web3.eth.getTransactionReceipt, txn_hash),
        success_fn=lambda r: r is not None and r['blockHash'] is not None,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )


def wait_for_block_number(web3, block_number=1, timeout=120, poll_interval=None):
    if is_tester_web3(web3):
        if hasattr(web3, 'manager'):
            rm = web3.manager
        else:
            rm = web3._requestManager

        while web3.eth.blockNumber < block_number:
            rm.request_blocking("evm_mine", [])
        return web3.eth.getBlock(block_number)
    return poll_until(
        poll_fn=lambda: web3.eth.blockNumber,
        success_fn=lambda v: v >= block_number,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )


def wait_for_unlock(web3, account=None, timeout=120, poll_interval=None):
    from .accounts import is_account_locked

    if account is None:
        account = web3.eth.coinbase

    return poll_until(
        poll_fn=functools.partial(is_account_locked, web3, account),
        success_fn=lambda v: v,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )


def wait_for_peers(web3, peer_count=1, timeout=120, poll_interval=None):
    return poll_until(
        poll_fn=lambda: web3.net.peerCount,
        success_fn=lambda v: v >= peer_count,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )


def wait_for_syncing(web3, timeout=120, poll_interval=None):
    return poll_until(
        poll_fn=lambda: web3.eth.syncing,
        success_fn=lambda v: v,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )


def wait_for_popen(proc, timeout=5, poll_interval=None):
    return poll_until(
        poll_fn=lambda: proc.poll,
        success_fn=lambda v: v,
        timeout=timeout,
        poll_interval_fn=lambda: poll_interval if poll_interval is not None else random.random(),
    )



from __future__ import absolute_import

from eth_utils import (
    to_dict,
)

from populus.utils.contracts import (
    is_project_contract,
)

from .base import (
    BaseContractBackend,
)


class ProjectContractsBackend(BaseContractBackend):
    """
    Provides access to compiled contract assets sources from the project
    `contracts_source_dirs`
    """
    is_provider = True
    is_registrar = False

    #
    # Provider API
    #
    def get_contract_identifier(self, contract_name):
        return contract_name

    @to_dict
    def get_all_contract_data(self):
        compiled_contracts = self.chain.project.compiled_contract_data
        for contract_name, contract_data in compiled_contracts.items():
            if is_project_contract(self.chain.project.contracts_source_dirs, contract_data):
                yield contract_name, contract_data



import logging
import sys
import warnings

import click

from populus.utils.filesystem import (
    is_same_path,
)
from populus.project import (
    Project,
)
from populus.config.versions import (
    LATEST_VERSION,
)

from populus.utils.logging import (
    get_logger_with_click_handler,
)


CONTEXT_SETTINGS = dict(
    # Support -h as a shortcut for --help
    help_option_names=['-h', '--help'],
)


def validate_logging_level(ctx, param, value):
    normalized_value = value.lower()
    if normalized_value.lower() in {'info', '20'}:
        return logging.INFO
    elif normalized_value.lower() in {'debug', '10'}:
        return logging.DEBUG
    else:
        raise click.BadParameter(
            "Logging level must be one of DEBUG/INFO or the numeric equivalents "
            "10/20"
        )


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    '--project',
    '-p',
    'project_dir',
    help=(
        "Specify a populus project directory"
    ),
    type=click.Path(exists=True, dir_okay=True),
)
@click.option(
    '--logging',
    '-l',
    'logging_level',
    help=(
        "Specify the logging level.  Allowed values are DEBUG/INFO or their "
        "numeric equivalents 10/20"
    ),
    default=str(logging.INFO),
    callback=validate_logging_level,
)
@click.pass_context
def main(ctx, project_dir, logging_level):
    """
    Populus
    """
    logger = get_logger_with_click_handler('populus', level=logging_level)
    ctx.obj = {}
    ctx.obj['PROJECT_DIR'] = project_dir

    if ctx.invoked_subcommand not in ('init', 'upgrade'):

        project = Project(project_dir)

        config_version = project.config['version']
        subcommand_bypasses_config_version = ctx.invoked_subcommand in {'config'}

        if not subcommand_bypasses_config_version and config_version != LATEST_VERSION:
            old_config_version_msg = (
                "================ warning =================\n"
                "Your populus config file is current at version {0}. "
                "The latest version is {1}.  You can use the `populus config "
                "upgrade` command to upgrade your config file to the latest version\n"
                "================ warning =================\n\n".format(
                    config_version,
                    LATEST_VERSION,
                )
            )
            warnings.warn(DeprecationWarning(old_config_version_msg))
            logger.warning(old_config_version_msg)
            proceed_msg = (
                "Without and up-to-date configuration file Populus may not function "
                "correctly.  Would you still like to proceed?"
            )
            if not click.confirm(proceed_msg):
                ctx.exit(1)

        if not any(is_same_path(p, project.project_dir) for p in sys.path):
            # ensure that the project directory is in the sys.path
            sys.path.insert(0, project.project_dir)

        ctx.obj['PROJECT'] = project



from __future__ import absolute_import

import copy
import itertools
import pprint

from eth_utils import (
    is_dict,
    is_list_like,
)

from populus.utils.mappings import (
    set_nested_key,
    get_nested_key,
    has_nested_key,
    pop_nested_key,
    deep_merge_dicts,
)

from populus.config.defaults import (
    load_default_config,
)
from populus.config.validation import (
    get_validation_errors,
    format_errors,
)
from populus.config.versions import (
    V4,
    V5,
)


NEW_V5_PATHS = {
    'compilation.backends.SolcAutoBackend',
    'compilation.backends.SolcStandardJSON',
}

MOVED_V4_PATHS = {
}

MODIFIED_V4_PATHS = {
    'compilation.backends.SolcCombinedJSON.settings.output_values',
}


def upgrade_v4_to_v5(v4_config):
    """
    Upgrade a v4 config file to a v5 config file.
    """
    errors = get_validation_errors(v4_config, version=V4)
    if errors:
        raise ValueError(
            "Cannot upgrade invalid config.  Please ensure that your current "
            "configuration file is valid:\n\n{0}".format(
                format_errors(errors),
            )
        )

    v4_default_config = load_default_config(version=V4)
    v5_default_config = load_default_config(version=V5)

    if v4_config == v4_default_config:
        return v5_default_config

    upgraded_v4_config = copy.deepcopy(v4_config)

    # new configuration values whos keys were not present in the previous
    # configuration.
    for key_path in NEW_V5_PATHS:
        if has_nested_key(upgraded_v4_config, key_path):
            continue
        set_nested_key(
            upgraded_v4_config,
            key_path,
            get_nested_key(v5_default_config, key_path),
        )

    # keys in the new configuration that were relocated.
    for old_path, new_path in MOVED_V4_PATHS.items():
        default_value = get_nested_key(v5_default_config, new_path)

        if has_nested_key(upgraded_v4_config, old_path):
            existing_value = pop_nested_key(upgraded_v4_config, old_path)

            if is_dict(default_value) and is_dict(existing_value):
                merged_value = deep_merge_dicts(default_value, existing_value)
            elif is_list_like(default_value) and is_list_like(existing_value):
                merged_value = list(set(itertools.chain(default_value, existing_value)))
            else:
                raise ValueError(
                    "Unable to merge {0} with {1}".format(
                        type(default_value),
                        type(existing_value),
                    )
                )

            set_nested_key(
                upgraded_v4_config,
                new_path,
                merged_value,
            )
        else:
            set_nested_key(
                upgraded_v4_config,
                new_path,
                default_value,
            )

    # keys from the previous configuration that were changed.
    for key_path in MODIFIED_V4_PATHS:
        new_default = get_nested_key(v5_default_config, key_path)
        if key_path not in upgraded_v4_config:
            set_nested_key(
                upgraded_v4_config,
                key_path,
                new_default,
            )
        else:
            current_value = get_nested_key(upgraded_v4_config, key_path)
            old_default = get_nested_key(v4_default_config, key_path)
            if current_value == old_default:
                set_nested_key(
                    upgraded_v4_config,
                    key_path,
                    new_default,
                )

    # bump the version
    set_nested_key(upgraded_v4_config, 'version', V5)

    errors = get_validation_errors(upgraded_v4_config, version=V5)
    if errors:
        raise ValueError(
            "Upgraded configuration did not pass validation:\n\n"
            "\n=============Original-Configuration============\n"
            "{0}"
            "\n=============Upgraded-Configuration============\n"
            "{1}"
            "\n=============Validation-Errors============\n"
            "{2}".format(
                pprint.pformat(dict(v4_config)),
                pprint.pformat(dict(upgraded_v4_config)),
                format_errors(errors),
            )
        )

    return upgraded_v4_config



from __future__ import absolute_import

from eth_utils import (
    to_dict,
)

from web3.contract import (
    Contract,
)

from populus.utils.contracts import (
    is_project_contract,
    is_test_contract,
)
from populus.utils.functional import (
    to_object,
)


class PopulusContract(Contract):
    populus_meta = None


@to_object('PopulusMeta')
@to_dict
def build_populus_meta(chain, contract_data):
    yield (
        'is_project_contract',
        is_project_contract(chain.project.contracts_source_dirs, contract_data),
    )
    yield (
        'is_test_contract',
        is_test_contract(chain.project.tests_dir, contract_data),
    )
    yield 'contract_type_name', contract_data['name']
    yield 'source_path', contract_data['source_path']


CONTRACT_FACTORY_FIELDS = {
    'abi',
    'asm',
    'ast',
    'bytecode',
    'bytecode_runtime',
    'clone_bin',
    'dev_doc',
    'interface',
    'metadata',
    'opcodes',
    'src_map',
    'src_map_runtime',
    'user_doc',
}


def construct_contract_factory(chain, contract_identifier, contract_data):
    factory_kwargs = {
        key: contract_data[key]
        for key
        in CONTRACT_FACTORY_FIELDS
        if key in contract_data
    }
    populus_meta = build_populus_meta(chain, contract_data)
    return chain.web3.eth.contract(
        ContractFactoryClass=PopulusContract,
        populus_meta=populus_meta,
        **factory_kwargs
    )



import contextlib
import errno
import fnmatch
import functools
import os
import shutil
import sys
import tempfile as _tempfile

from eth_utils import (
    to_tuple,
    is_string,
    is_list_like,
)


if sys.version_info.major == 2:
    FileNotFoundError = OSError


def ensure_path_exists(dir_path):
    """
    Make sure that a path exists
    """
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
        return True
    return False


def ensure_file_exists(file_path):
    """
    Make sure that a path exists
    """
    if os.path.exists(file_path):
        return False
    base_dir = os.path.dirname(file_path)
    ensure_path_exists(base_dir)
    with open(file_path, 'w'):
        pass
    return True


def remove_file_if_exists(path):
    if os.path.isfile(path):
        os.remove(path)
        return True
    return False


def remove_dir_if_exists(path):
    if os.path.isdir(path):
        shutil.rmtree(path)
        return True
    return False


def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def is_executable_available(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath = os.path.dirname(program)
    if fpath:
        if is_exe(program):
            return True
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return True

    return False


@to_tuple
def recursive_find_files(base_dir, pattern):
    """
    Recursively traverse directory tree and find files matching pattern.

    :param base_dir: Base directory to start crawling.
    :param pattern: File pattern, to match again.
                    Either string (single pattern match) or tuple (multiple patterns)
    """

    def match(filename, pattern):
        if is_list_like(pattern):
            return any([fnmatch.fnmatch(filename, p) for p in pattern])
        elif is_string(pattern):
            return fnmatch.fnmatch(filename, pattern)
        else:
            raise TypeError(
                "Pattern must either be a string pattern or a list of patterns." +
                "  Got {0}".format(pattern)
            )

    for dirpath, _, filenames in os.walk(base_dir):
        for filename in filenames:
            if match(filename, pattern):
                yield os.path.join(dirpath, filename)


@contextlib.contextmanager
def tempdir(*args, **kwargs):
    directory = _tempfile.mkdtemp(*args, **kwargs)

    try:
        yield directory
    finally:
        remove_dir_if_exists(directory)


@contextlib.contextmanager
def tempfile(*args, **kwargs):
    _, file_path = _tempfile.mkstemp(*args, **kwargs)

    try:
        yield file_path
    finally:
        remove_file_if_exists(file_path)


def is_same_path(p1, p2):
    n_p1 = os.path.abspath(os.path.expanduser(p1))
    n_p2 = os.path.abspath(os.path.expanduser(p2))

    try:
        return os.path.samefile(n_p1, n_p2)
    except FileNotFoundError:
        return n_p1 == n_p2


def relpath(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        value = fn(*args, **kwargs)
        if is_string(value):
            return os.path.relpath(value)
        elif is_list_like(value):
            return type(value)([
                os.path.relpath(path) for path in value
            ])
        else:
            raise TypeError("Unsupported type: {0}".format(type(value)))
    return wrapper


def normpath(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        path = fn(*args, **kwargs)
        return os.path.normpath(path)
    return wrapper


def is_under_path(base_path, path):
    if is_same_path(base_path, path):
        return False
    absolute_base_path = os.path.abspath(base_path)
    absolute_path = os.path.abspath(path)
    return absolute_path.startswith(absolute_base_path)


def get_latest_mtime(file_paths):
    if not file_paths:
        return None
    else:
        return max(
            os.path.getmtime(path)
            for path
            in file_paths
        )



import functools
import itertools

from eth_utils import (
    to_tuple,
)

from populus.utils.contracts import (
    EMPTY_BYTECODE_VALUES,
    find_deploy_block_number,
)
from populus.utils.functional import chain_return

from .exceptions import (
    NoKnownAddress,
)


def address_sort_fn(web3, address):
    try:
        return find_deploy_block_number(web3, address)
    except ValueError:
        return -1


class Registrar(object):
    """
    Abstraction for recording known contracts on a given chain.
    """
    registrar_backends = None

    def __init__(self, chain, registrar_backends):
        self.chain = chain
        self.registrar_backends = registrar_backends

    def set_contract_address(self, contract_name, contract_address):
        """
        Set a contract address in the registrar
        """
        return [
            registrar.set_contract_address(contract_name, contract_address)
            for registrar in self.registrar_backends.values()
        ]

    def get_contract_addresses(self, contract_identifier):
        """
        Retrieve a contract address from the registrar
        """
        found_addresses = self._get_contract_addresses_from_backends(contract_identifier)
        if not found_addresses:
            raise NoKnownAddress("No known address for contract")

        if len(found_addresses) == 1:
            return found_addresses

        addresses_with_code = tuple(set(
            address
            for address
            in found_addresses
            if self.chain.web3.eth.getCode(address) not in EMPTY_BYTECODE_VALUES
        ))
        empty_addresses = tuple(set(
            address
            for address
            in found_addresses
            if self.chain.web3.eth.getCode(address) in EMPTY_BYTECODE_VALUES
        ))

        if len(addresses_with_code) > 1:
            sorted_addresses = tuple(sorted(
                addresses_with_code,
                key=functools.partial(address_sort_fn, self.chain.web3),
                reverse=True,
            ))
        else:
            sorted_addresses = addresses_with_code

        known_addresses = tuple(itertools.chain(
            sorted_addresses,
            empty_addresses,
        ))

        return known_addresses

    @to_tuple
    @chain_return
    def _get_contract_addresses_from_backends(self, contract_identifier):
        for registrar in self.registrar_backends.values():
            try:
                yield registrar.get_contract_addresses(contract_identifier)
            except NoKnownAddress:
                continue



import sys

from eth_utils import (
    force_bytes,
    force_text,
)


def normalize_class_name(value):
    """
    For `type()` calls:
    * Python 2 wants `str`
    * Python 3.4 wants `str`
    * Python 3.5 doesn't care.
    """
    if sys.version_info.major == 2:
        return force_bytes(value)
    else:
        return force_text(value)



from __future__ import absolute_import

import json
import os

from eth_utils import (
    to_tuple,
)

from populus.contracts.exceptions import (
    NoKnownAddress,
)

from populus.utils.mappings import (
    set_nested_key,
)
from populus.utils.chains import (
    get_chain_definition,
    check_if_chain_matches_chain_uri,
)

from .base import (
    BaseContractBackend,
)


@to_tuple
def get_matching_chain_definitions(web3, values):
    for chain_definition in values:
        if check_if_chain_matches_chain_uri(web3, chain_definition):
            yield chain_definition


def load_registrar_data(registrar_file):
    registrar_data = json.load(registrar_file)
    registrar_data.setdefault('deployments', {})

    return registrar_data


class JSONFileBackend(BaseContractBackend):
    is_registrar = True
    is_provider = False

    #
    # Registrar API
    #
    def set_contract_address(self, instance_identifier, address):
        registrar_data = self.registrar_data

        chain_definition = get_chain_definition(self.chain.web3)
        set_nested_key(
            registrar_data,
            'deployments.{0}.{1}'.format(chain_definition, instance_identifier),
            address,
        )

        with open(self.registrar_path, 'w') as registrar_file:
            json.dump(
                registrar_data,
                registrar_file,
                sort_keys=True,
                indent=2,
                separators=(',', ': '),
            )

    @to_tuple
    def get_contract_addresses(self, instance_identifier):
        registrar_data = self.registrar_data
        matching_chain_definitions = get_matching_chain_definitions(
            self.chain.web3,
            registrar_data.get('deployments', {}),
        )

        identifier_exists = any(
            instance_identifier in registrar_data['deployments'][chain_definition]
            for chain_definition
            in matching_chain_definitions
        )
        if not identifier_exists:
            raise NoKnownAddress("No known address for '{0}'".format(instance_identifier))

        for chain_definition in matching_chain_definitions:
            chain_deployments = registrar_data['deployments'][chain_definition]
            if instance_identifier in chain_deployments:
                yield chain_deployments[instance_identifier]

    #
    # Private API
    #
    @property
    def registrar_path(self):
        file_path = self.config.get('file_path', './registrar.json')
        return os.path.join(self.chain.project.project_dir, file_path)

    @property
    def registrar_data(self):
        # TODO: this could be cached using mtime of the file.
        if os.path.exists(self.registrar_path):
            with open(self.registrar_path, 'r') as registrar_file:
                registrar_data = load_registrar_data(registrar_file)
        else:
            registrar_data = {'deployments': {}}

        return registrar_data



from populus.contracts.exceptions import (  # noqa: F401
    NoKnownAddress,
    UnknownContract,
    BytecodeMismatch,
)

from .geth import (  # noqa: F401
    BaseGethChain,
    LocalGethChain,
    MainnetChain,
    TemporaryGethChain,
    TestnetChain,
)
from .external import (  # noqa: F401
    ExternalChain,
)
from .tester import (  # noqa: F401
    TesterChain,
)
from .testrpc import (  # noqa: F401
    TestRPCChain,
)


__all__ = (
    "TestRPCChain",
    "TesterChain",
    "BaseGethChain",
    "TemporaryGethChain",
    "LocalGethChain",
    "MainnetChain",
    "TestnetChain",
    "ExternalChain",
)



from .base import (  # noqa: F401
    Config,
)
from .web3 import (  # noqa: F401
    Web3Config,
)
from .chain import (  # noqa: F401
    ChainConfig,
)
from .compiler import (  # noqa: F401
    CompilerConfig,
)
from .loading import (  # noqa: F401
    load_config,
    write_config,
)
from .defaults import (  # noqa: F401
    load_default_config,
)
from .validation import (  # noqa: F401
    load_config_schema,
    validate_config,
)



import itertools

from pylru import lrucache

from eth_utils import (
    to_tuple,
)

from populus.utils.contracts import (
    verify_contract_bytecode,
)
from populus.utils.linking import (
    link_bytecode,
)

from .exceptions import (
    BytecodeMismatch,
    NoKnownAddress,
    UnknownContract,
)


def get_base_contract_factory(contract_identifier, store_backends):
    for backend in store_backends.values():
        try:
            return backend.get_base_contract_factory(contract_identifier)
        except UnknownContract:
            pass
    else:
        raise UnknownContract(
            "No contract data was available for the contract identifier '{0}' "
            "from any of the configured backends".format(contract_identifier)
        )


@to_tuple
def filter_addresses_by_bytecode_match(web3, expected_bytecode, addresses):
    for address in addresses:
        try:
            verify_contract_bytecode(web3, expected_bytecode, address)
        except BytecodeMismatch:
            continue
        else:
            yield address


class Provider(object):
    """
    Abstraction for retrieving contracts on a given chain.
    """
    provider_backends = None

    def __init__(self, chain, provider_backends):
        self.chain = chain
        self.provider_backends = provider_backends
        self._factory_cache = lrucache(128)

    def is_contract_available(self, contract_identifier):
        try:
            contract_addresses = self.chain.registrar.get_contract_addresses(contract_identifier)
        except NoKnownAddress:
            return False

        if not self.are_contract_dependencies_available(contract_identifier):
            return False

        ContractFactory = self.get_contract_factory(contract_identifier)
        bytecode_matched_addresses = filter_addresses_by_bytecode_match(
            self.chain.web3,
            ContractFactory.bytecode_runtime,
            contract_addresses,
        )

        if not bytecode_matched_addresses:
            return False

        return True

    def are_contract_dependencies_available(self, contract_identifier):
        contract_data = self.get_contract_data(contract_identifier)
        contract_dependencies = contract_data['ordered_full_dependencies']

        return all(self.is_contract_available(c) for c in contract_dependencies)

    def get_contract(self, contract_identifier):
        ContractFactory = self.get_contract_factory(contract_identifier)
        contract_addresses = self.chain.registrar.get_contract_addresses(contract_identifier)

        bytecode_matched_addresses = filter_addresses_by_bytecode_match(
            self.chain.web3,
            ContractFactory.bytecode_runtime,
            contract_addresses,
        )
        if not bytecode_matched_addresses:
            raise BytecodeMismatch("None of the known addresses matched the expected bytecode")
        else:
            # TODO: don't just default to the first address.
            contract_address = contract_addresses[0]

        return ContractFactory(address=contract_address)

    def deploy_contract(self,
                        contract_identifier,
                        deploy_transaction=None,
                        deploy_args=None,
                        deploy_kwargs=None):
        """
        Same as get_contract but it will also lazily deploy the contract with
        the provided deployment arguments
        """
        contract_data = self.get_contract_data(contract_identifier)
        contract_dependencies = contract_data['ordered_full_dependencies']

        for dependency_name in contract_dependencies:
            self.get_or_deploy_contract(dependency_name, deploy_transaction=deploy_transaction)

        ContractFactory = self.get_contract_factory(contract_identifier)
        deploy_transaction_hash = ContractFactory.deploy(
            transaction=deploy_transaction,
            args=deploy_args,
            kwargs=deploy_kwargs,
        )
        contract_address = self.chain.wait.for_contract_address(deploy_transaction_hash)
        registrar = self.chain.registrar
        registrar.set_contract_address(contract_identifier, contract_address)

        return self.get_contract(contract_identifier), deploy_transaction_hash

    def get_or_deploy_contract(self,
                               contract_identifier,
                               deploy_transaction=None,
                               deploy_args=None,
                               deploy_kwargs=None):
        """
        Same as get_contract but it will also lazily deploy the contract with
        the provided deployment arguments
        """
        if not self.is_contract_available(contract_identifier):
            return self.deploy_contract(
                contract_identifier=contract_identifier,
                deploy_transaction=deploy_transaction,
                deploy_args=deploy_args,
                deploy_kwargs=deploy_kwargs,
            )
        else:
            return (
                self.get_contract(contract_identifier),
                None,
            )

    def get_base_contract_factory(self, contract_identifier):
        """
        Returns the base contract factory for the given `contract_identifier`.
        The `bytecode` and `bytecode_runtime` will be unlinked in this class.
        """
        return get_base_contract_factory(contract_identifier, self.provider_backends)

    def get_contract_data(self, contract_identifier):
        """
        Returns a dictionary containing the compiler output for the given
        contract identifier.
        """
        for backend in self.provider_backends.values():
            try:
                return backend.get_contract_data(contract_identifier)
            except UnknownContract:
                continue
        else:
            raise UnknownContract(
                "No contracts found for the contract identifier '{0}'".format(
                    contract_identifier,
                )
            )

    def get_all_contract_data(self):
        """
        Returns a dictionary mapping all contract identifiers to their contract data.
        """
        return dict(itertools.chain.from_iterable(
            backend.get_all_contract_data().items()
            for backend
            in self.provider_backends.values()
        ))

    def get_all_contract_names(self):
        """
        Returns a set of all of the known contract identifiers.
        """
        return set(itertools.chain.from_iterable(
            backend.get_all_contract_names()
            for backend
            in self.provider_backends.values()
        ))

    def get_contract_factory(self, contract_identifier):
        """
        Returns the contract factory for the given `contract_identifier`.  The
        `bytecode` and `bytecode_runtime` values for this factory will be fully
        linked.
        """
        if contract_identifier in self._factory_cache:
            return self._factory_cache[contract_identifier]

        BaseContractFactory = self.get_base_contract_factory(contract_identifier)

        contract_data = self.get_contract_data(contract_identifier)

        bytecode = self._link_bytecode(contract_data['bytecode'], contract_data['linkrefs'])
        bytecode_runtime = self._link_bytecode(
            contract_data['bytecode_runtime'],
            contract_data['linkrefs_runtime'],
        )

        ContractFactory = BaseContractFactory.factory(
            web3=BaseContractFactory.web3,
            bytecode=bytecode,
            bytecode_runtime=bytecode_runtime,
        )

        self._factory_cache[contract_identifier] = ContractFactory
        return ContractFactory

    #
    # Private API
    #
    def _link_bytecode(self, bytecode, link_references):
        """
        Return the fully linked contract bytecode.

        Note: This *must* use `get_contract` and **not** `get_contract_address`
        for resolution of link dependencies.  If it merely uses
        `get_contract_address` then the bytecode of sub-dependencies is not
        verified.
        """
        resolved_link_references = tuple(
            (
                link_reference,
                self.get_contract(link_reference['name']).address
            )
            for link_reference
            in link_references
        )

        linked_bytecode = link_bytecode(bytecode, resolved_link_references)

        return linked_bytecode



from .base import (
    BaseChain,
)


class TesterChain(BaseChain):
    def __enter__(self):
        if self._running:
            raise ValueError("The TesterChain is already running")

        self._running = True

        self.rpc_methods = self.web3.providers[0].rpc_methods

        self.rpc_methods.full_reset()
        self.rpc_methods.rpc_configure('eth_mining', False)
        self.rpc_methods.rpc_configure('eth_protocolVersion', '0x3f')
        self.rpc_methods.rpc_configure('net_version', 1)
        self.rpc_methods.evm_mine()

        return self



import logging
import click
from .main import main

from populus.api.project import (
    init_project,
)


@main.command('init')
@click.pass_context
def init_cmd(ctx):
    """
    Generate project layout with an example contract.
    """
    logger = logging.getLogger('populus.cli.init_cmd')
    project_dir = ctx.obj['PROJECT_DIR']

    init_project(project_dir, logger)



def test_greeter(chain):
    greeter, _ = chain.provider.get_or_deploy_contract('Greeter')

    greeting = greeter.call().greet()
    assert greeting == 42


def test_custom_greeting(chain):
    greeter, _ = chain.provider.get_or_deploy_contract('Greeter')

    set_txn_hash = greeter.transact().setGreeting(1337)
    chain.wait.for_receipt(set_txn_hash)

    greeting = greeter.call().greet()
    assert greeting == 1337



import itertools
import re

from eth_utils import (
    remove_0x_prefix,
    to_dict,
)

from .filesystem import (
    is_under_path,
)
from .string import (
    normalize_class_name,
)


def is_project_contract(contracts_source_dirs, contract_data):
    return any(
        is_under_path(source_dir, contract_data['source_path'])
        for source_dir
        in contracts_source_dirs
    )


def is_test_contract(tests_dir, contract_data):
    return is_under_path(tests_dir, contract_data['source_path'])


def package_contracts(contract_factories):
    _dict = {
        '__len__': lambda s: len(contract_factories),
        '__iter__': lambda s: iter(contract_factories.items()),
        '__contains__': lambda s, k: contract_factories.__contains__(k),
        '__getitem__': lambda s, k: contract_factories.__getitem__(k),
        '__setitem__': lambda s, k, v: contract_factories.__setitem__(k, v),
        'keys': lambda s: contract_factories.keys(),
        'values': lambda s: contract_factories.values(),
    }
    _dict.update(contract_factories)

    return type('contracts', (object,), _dict)()


CONTRACT_FACTORY_FIELDS = {
    'abi',
    'asm',
    'ast',
    'bytecode',
    'bytecode_runtime',
    'clone_bin',
    'dev_doc',
    'interface',
    'metadata',
    'opcodes',
    'src_map',
    'src_map_runtime',
    'user_doc',
}


def create_contract_factory(web3, contract_name, contract_data):
    factory_kwargs = {
        key: contract_data[key]
        for key
        in CONTRACT_FACTORY_FIELDS
        if key in contract_data
    }
    return web3.eth.contract(
        contract_name=normalize_class_name(contract_name),
        **factory_kwargs
    )


def construct_contract_factories(web3, compiled_contracts):
    contract_classes = {
        contract_name: create_contract_factory(
            web3,
            contract_name,
            contract_data,
        )
        for contract_name, contract_data
        in compiled_contracts.items()
    }

    return package_contracts(contract_classes)


@to_dict
def compute_direct_dependency_graph(compiled_contracts):
    """
    Given a dictionary or mapping of compiled contract data, this returns a *shallow*
    dependency graph of each contracts explicit link dependencies.
    """
    for contract_data in compiled_contracts:
        yield (
            contract_data['name'],
            contract_data['direct_dependencies'],
        )


def compute_recursive_contract_dependencies(contract_name, dependency_graph):
    """
    Recursive computation of the linker dependencies for a specific contract
    within a contract dependency graph.
    """
    direct_dependencies = dependency_graph.get(contract_name, set())
    sub_dependencies = itertools.chain.from_iterable((
        compute_recursive_contract_dependencies(dep, dependency_graph)
        for dep in direct_dependencies
    ))
    return set(itertools.chain(direct_dependencies, sub_dependencies))


CONTRACT_NAME_REGEX = '^[_a-zA-Z][_a-zA-Z0-9]*$'


def is_contract_name(value):
    return bool(re.match(CONTRACT_NAME_REGEX, value))


EMPTY_BYTECODE_VALUES = {None, "0x"}


SWARM_HASH_PREFIX = "a165627a7a72305820"
SWARM_HASH_SUFFIX = "0029"
EMBEDDED_SWARM_HASH_REGEX = (
    SWARM_HASH_PREFIX +
    "[0-9a-zA-Z]{64}" +
    SWARM_HASH_SUFFIX +
    "$"
)

SWARM_HASH_REPLACEMENT = (
    SWARM_HASH_PREFIX +
    "<" +
    "-" * 20 +
    "swarm-hash-placeholder" +
    "-" * 20 +
    ">" +
    SWARM_HASH_SUFFIX
)


def compare_bytecode(left, right):
    unprefixed_left = remove_0x_prefix(left)
    unprefixed_right = remove_0x_prefix(right)

    norm_left = re.sub(EMBEDDED_SWARM_HASH_REGEX, SWARM_HASH_REPLACEMENT, unprefixed_left)
    norm_right = re.sub(EMBEDDED_SWARM_HASH_REGEX, SWARM_HASH_REPLACEMENT, unprefixed_right)

    if len(norm_left) != len(unprefixed_left) or len(norm_right) != len(unprefixed_right):
        raise ValueError(
            "Invariant.  Normalized bytecodes are not the correct lengths:" +
            "\n- left  (original)  :" +
            left, +
            "\n- left  (unprefixed):" +
            unprefixed_left +
            "\n- left  (normalized):" +
            norm_left +
            "\n- right (original)  :" +
            right +
            "\n- right (unprefixed):" +
            unprefixed_right +
            "\n- right (normalized):" +
            norm_right
        )

    return norm_left == norm_right


def verify_contract_bytecode(web3, expected_bytecode, address):
    """
    TODO: write tests for this.
    """
    from populus.contracts.exceptions import BytecodeMismatch

    # Check that the contract has bytecode
    if expected_bytecode in EMPTY_BYTECODE_VALUES:
        raise ValueError(
            "Contract instances which contain an address cannot have empty "
            "runtime bytecode"
        )

    chain_bytecode = web3.eth.getCode(address)

    if chain_bytecode in EMPTY_BYTECODE_VALUES:
        raise BytecodeMismatch(
            "No bytecode found at address: {0}".format(address)
        )
    elif not compare_bytecode(chain_bytecode, expected_bytecode):
        raise BytecodeMismatch(
            "Bytecode found at {0} does not match compiled bytecode:\n"
            " - chain_bytecode: {1}\n"
            " - compiled_bytecode: {2}".format(
                address,
                chain_bytecode,
                expected_bytecode,
            )
        )


def find_deploy_block_number(web3, address):
    chain_bytecode = web3.eth.getCode(address, "latest")
    if chain_bytecode in EMPTY_BYTECODE_VALUES:
        raise NotImplementedError("Cannot find deploy transaction for address with empty code")

    left = 0
    right = web3.eth.blockNumber

    while left + 1 < right:
        middle = (left + right) // 2
        # This only works if the node was not fast synced for the provided
        # `block_identifier`.
        try:
            middle_code = web3.eth.getCode(address, block_identifier=middle)
        except ValueError as err:
            if 'Missing trie node' in str(err):
                left = middle
                continue
            raise

        if middle_code in EMPTY_BYTECODE_VALUES:
            left = middle
        else:
            right = middle

    code_at_right = web3.eth.getCode(address, block_identifier=right)
    if code_at_right in EMPTY_BYTECODE_VALUES:
        raise ValueError(
            "Something went wrong with the binary search to find the deploy block"
        )
    code_at_previous_block = web3.eth.getCode(address, block_identifier=right - 1)
    if code_at_previous_block not in EMPTY_BYTECODE_VALUES:
        raise ValueError(
            "Something went wrong with the binary search to find the deploy block"
        )
    return right



import copy
import itertools
import os
import shutil
import sys
import warnings

from eth_utils import (
    to_tuple,
)

from populus.config.defaults import (
    load_default_config,
)

from populus.exceptions import (
    PopulusResourceWarning,
)

from populus.compilation import (
    compile_project_contracts,
)

from populus.config import (
    ChainConfig,
    CompilerConfig,
    Config,
    load_config as _load_config,
    load_config_schema,
    write_config,
)

from populus.config.defaults import (
    get_default_config_path,
    get_user_default_config_path,
)

from populus.config.helpers import (
    check_if_user_json_config_file_exists,
    get_user_json_config_file_path,
)

from populus.utils.chains import (
    get_base_blockchain_storage_dir,
)

from populus.utils.compile import (
    get_build_asset_dir,
    get_compiled_contracts_asset_path,
    get_contracts_source_dirs,
)

from populus.utils.filesystem import (
    ensure_path_exists,
    get_latest_mtime,
)

from populus.config.helpers import (
    get_json_config_file_path,
    get_legacy_json_config_file_path,
)

from populus.utils.testing import (
    get_tests_dir,
)


if sys.version_info.major == 2:
    FileNotFoundError = OSError


class Project(object):

    project_dir = None
    config_file_path = None
    user_config_file_path = None
    legacy_config_path = None

    def __init__(self, project_dir=None, user_config_file_path=None, create_config_file=False):  # noqa: E501

        self._reset_configs_cache()
        if project_dir is None:
            self.project_dir = os.getcwd()
        else:
            self.project_dir = os.path.abspath(project_dir)

        # user config
        self.user_config_file_path = user_config_file_path
        if self.user_config_file_path is None:
            self.user_config_file_path = get_user_json_config_file_path()
            if not check_if_user_json_config_file_exists():
                user_config_path = get_user_json_config_file_path()
                user_defaults_path = get_user_default_config_path()
                ensure_path_exists(os.path.dirname(user_config_path))
                shutil.copyfile(
                    user_defaults_path,
                    user_config_path
                )

        # legacy config
        legacy_path = get_legacy_json_config_file_path(self.project_dir)
        if os.path.exists(legacy_path):
            self.legacy_config_path = legacy_path
            warnings.warn(
                "Found legacy config file at {legacy_path}".format(
                    legacy_path=legacy_path
                ),
                PopulusResourceWarning
            )

        # project config
        self.config_file_path = get_json_config_file_path(self.project_dir)
        if not os.path.exists(self.config_file_path):
            if create_config_file:
                defaults_path = get_default_config_path()
                shutil.copyfile(defaults_path, self.config_file_path)
            else:
                if self.legacy_config_path is not None:
                    msg = "No project config file found at {project_dir}, but a legacy config exists. Try to upgrade"  # noqa: E501
                else:
                    msg = "No project config file found at {project_dir}"
                raise FileNotFoundError(
                    msg.format(project_dir=self.project_dir)
                )

        self.load_config()

    #
    # Config
    #

    _project_config = None
    _user_config = None
    _config_schema = None

    def load_config(self):
        self._reset_configs_cache()
        self._project_config = _load_config(self.config_file_path)
        self._user_config = _load_config(self.user_config_file_path)

        config_version = self._project_config['version']
        self._config_schema = load_config_schema(config_version)

    def _reset_configs_cache(self):
        self._merged_config_cache = None
        self._user_config_cache = None
        self._project_config_cache = None

    def reload_config(self):
        self.load_config()

    @property
    def user_config(self):

        if self._user_config_cache is None:
            user_config = Config(
                config=self._user_config,
                schema=self._config_schema,
            )
            user_config.unref()
            self._user_config_cache = user_config
            self.merge_user_and_project_configs(user_config, self.project_config)

        return self._user_config_cache

    @property
    def project_config(self):
        if self._project_config_cache is None:
            project_config = Config(
                config=self._project_config,
                parent=Config(self._user_config)
            )
            project_config.unref()
            self._project_config_cache = project_config

            # schema validation
            # partial project config must be merged to get the entire schema
            self.merge_user_and_project_configs(self.user_config, project_config)

        return self._project_config_cache

    def merge_user_and_project_configs(self, user_config, project_config):
        if self._merged_config_cache is None:
            merged_config = copy.deepcopy(self.user_config)
            for key, value in self.project_config.items(flatten=True):
                if key != 'version':
                    merged_config[key] = value
            Config(config=merged_config, schema=self._config_schema)
            self._merged_config_cache = merged_config

        return self._merged_config_cache

    @property
    def config(self):

        return self.merge_user_and_project_configs(self.user_config, self.project_config)

    @config.setter
    def config(self, value):
        if isinstance(value, Config):
            self._merged_config_cache = value
        else:
            self._merged_config_cache = Config(
                config=value,
                schema=self._config_schema
            )

    def clean_config(self):

        items = self.project_config.items(flatten=True)

        default_config = Config(load_default_config(version=self.project_config['version']))
        default_config.unref()
        default_project_keys = [x[0] for x in default_config.items(flatten=True)]

        for key, value in items:
            if self.user_config.get(key) == value and key not in default_project_keys:
                self.project_config.pop(key)

        write_config(self.project_config, self.config_file_path)

    #
    # Project
    #
    @property
    def tests_dir(self):
        return get_tests_dir(self.project_dir)

    #
    # Contracts
    #
    @property
    def compiled_contracts_asset_path(self):
        return get_compiled_contracts_asset_path(self.build_asset_dir)

    @property
    @to_tuple
    def contracts_source_dirs(self):
        source_dirs = self.config.get('compilation.contracts_source_dirs')
        if source_dirs:
            return [os.path.join(self.project_dir, contracts_dir) for contracts_dir in source_dirs]
        else:
            return get_contracts_source_dirs(self.project_dir)

    @property
    def build_asset_dir(self):
        return get_build_asset_dir(self.project_dir)

    _cached_compiled_contracts_mtime = None
    _cached_compiled_contracts = None

    @to_tuple
    def get_all_source_file_paths(self):
        compiler_backend = self.get_compiler_backend()
        return itertools.chain.from_iterable(
            compiler_backend.get_project_source_paths(source_dir)
            for source_dir
            in self.contracts_source_dirs
        )

    def is_compiled_contract_cache_stale(self):
        if self._cached_compiled_contracts is None:
            return True

        source_mtime = get_latest_mtime(self.get_all_source_file_paths())

        if source_mtime is None:
            return True
        elif self._cached_compiled_contracts_mtime is None:
            return True
        else:
            return self._cached_compiled_contracts_mtime < source_mtime

    def fill_contracts_cache(self, compiled_contracts, contracts_mtime):
        """
        :param contracts: become the Project's cache for compiled contracts
        :param contracts_mtime: last modification of supplied contracts
        :return:
        """
        self._cached_compiled_contracts_mtime = contracts_mtime
        self._cached_compiled_contracts = compiled_contracts

    @property
    def compiled_contract_data(self):
        if self.is_compiled_contract_cache_stale():
            source_file_paths, compiled_contracts = compile_project_contracts(self)
            contracts_mtime = get_latest_mtime(source_file_paths)
            self.fill_contracts_cache(
                compiled_contracts=compiled_contracts,
                contracts_mtime=contracts_mtime,
            )
        return self._cached_compiled_contracts

    #
    # Compiler Backend
    #
    def get_compiler_backend(self):
        compilation_config = self.config.get_config(
            'compilation.backend',
            config_class=CompilerConfig,
        )
        return compilation_config.backend

    #
    # Local Blockchains
    #
    def get_chain_config(self, chain_name):
        chain_config_key = 'chains.{chain_name}'.format(chain_name=chain_name)

        if chain_config_key in self.config:
            return self.config.get_config(chain_config_key, config_class=ChainConfig)
        else:
            raise KeyError(
                "Unknown chain: {0!r} - Must be one of {1!r}".format(
                    chain_name,
                    sorted(self.config.get('chains', {}).keys()),
                )
            )

    def get_chain(self, chain_name, chain_config=None):
        """
        Returns a context manager that runs a chain within the context of the
        current populus project.

        Alternatively you can specify any chain name that is present in the
        `chains` configuration key.
        """
        if chain_config is None:
            chain_config = self.get_chain_config(chain_name)
        chain = chain_config.get_chain(self, chain_name)
        return chain

    @property
    def base_blockchain_storage_dir(self):
        return get_base_blockchain_storage_dir(self.project_dir)



from __future__ import absolute_import

import itertools
import logging
import os

from populus.utils.compile import (
    validate_compiled_contracts,
    post_process_compiled_contracts,
)
from populus.utils.functional import (
    get_duplicates,
)


def _get_contract_key(contract_data):
    return (
        contract_data['source_path'],
        contract_data['name'],
    )


def compile_project_contracts(project):
    logger = logging.getLogger('populus.compilation.compile_project_contracts')
    compiler_backend = project.get_compiler_backend()

    project_contract_source_paths = tuple(itertools.chain.from_iterable(
        compiler_backend.get_project_source_paths(source_dir)
        for source_dir
        in project.contracts_source_dirs
    ))
    logger.debug(
        "Found %s project source files: %s",
        len(project_contract_source_paths),
        ", ".join(project_contract_source_paths),
    )

    test_contract_source_paths = compiler_backend.get_test_source_paths(project.tests_dir)
    logger.debug(
        "Found %s test source files: %s",
        len(test_contract_source_paths),
        ", ".join(test_contract_source_paths),
    )

    all_source_paths = tuple(itertools.chain(
        project_contract_source_paths,
        test_contract_source_paths,
    ))

    base_compiled_contracts = compiler_backend.get_compiled_contracts(
        source_file_paths=all_source_paths,
        import_remappings=project.config.get('compilation.import_remappings'),
    )
    compiled_contracts = post_process_compiled_contracts(base_compiled_contracts)
    validate_compiled_contracts(compiled_contracts)

    logger.info("> Found %s contract source files", len(all_source_paths))
    for path in sorted(all_source_paths):
        logger.info("  - %s", os.path.relpath(path))

    logger.info("> Compiled %s contracts", len(compiled_contracts))
    for contract_key in sorted(map(_get_contract_key, compiled_contracts)):
        logger.info("  - %s", ':'.join(contract_key))

    duplicate_contract_names = get_duplicates(
        contract_data['name']
        for contract_data
        in compiled_contracts
    )
    if duplicate_contract_names:
        raise ValueError(
            "Duplicate contract names '{0}'.  Populus cannot currently support "
            "compilation which produces multiple contracts with the same name".format(
                ", ".join(duplicate_contract_names),
            )
        )

    compiled_contracts_by_key = {
        contract_data['name']: contract_data
        for contract_data
        in compiled_contracts
    }

    return all_source_paths, compiled_contracts_by_key



import logging
import click
from .main import main

from populus.api.upgrade import (
    upgrade_configs,
)

from populus.config.versions import (
    LATEST_VERSION,
)


@main.command('upgrade')
@click.pass_context
@click.option(
    '-t',
    '--to-version',
    'to_version',
    default=LATEST_VERSION,
)
def upgrade_cmd(ctx, to_version):
    """
    Upgrade a project config, and if required also the user config
    Note: the user config is used in other projects as well
    """
    logger = logging.getLogger('populus.cli.upgrade')
    project_dir = ctx.obj['PROJECT_DIR']

    upgrade_configs(project_dir, logger)



from __future__ import absolute_import
import enum
import pprint

from cytoolz.functoolz import (
    pipe,
)

from eth_utils import (
    to_tuple,
)

from populus.config.versions import (
    V1,
    V2,
    V3,
    V4,
    V5,
    V6,
    V7,
    KNOWN_LEGACY_VERSIONS,
    KNOWN_USER_VERSIONS,
    LATEST_VERSION,
)
from .v1 import upgrade_v1_to_v2
from .v2 import upgrade_v2_to_v3
from .v3 import upgrade_v3_to_v4
from .v4 import upgrade_v4_to_v5
from .v5 import upgrade_v5_to_v6
from .v6 import upgrade_v6_to_v7


UPGRADE_SEQUENCE = {
    V1: V2,
    V2: V3,
    V3: V4,
    V4: V5,
    V5: V6,
    V6: V7,
}

UPGRADE_FUNCTIONS = {
    V1: upgrade_v1_to_v2,
    V2: upgrade_v2_to_v3,
    V3: upgrade_v3_to_v4,
    V4: upgrade_v4_to_v5,
    V5: upgrade_v5_to_v6,
    V6: upgrade_v6_to_v7,
}


class ConfigContext(enum.Enum):
    USER = 1
    LEGACY = 2


@to_tuple
def get_upgrade_sequence(start_version, end_version, known_versions):

    if start_version not in known_versions:
        raise KeyError("Unknown version '{0}':  Must be one of '{1}'".format(
            start_version,
            ', '.join(sorted(known_versions)),
        ))
    elif end_version not in known_versions:
        raise KeyError("Unknown version '{0}':  Must be one of '{1}'".format(
            end_version,
            ', '.join(sorted(known_versions)),
        ))
    elif start_version == end_version:
        raise ValueError("Config is already at version: '{0}'".format(end_version))

    elif int(start_version) > int(end_version):
        raise ValueError("Cannot downgrade config from version '{0}' to '{1}'".format(
            start_version,
            end_version,
        ))

    current_version = start_version
    while current_version != end_version:
        yield current_version
        current_version = UPGRADE_SEQUENCE[current_version]


def upgrade_config(config, config_context, to_version=LATEST_VERSION):

    if config_context == ConfigContext.USER:
        known_versions = KNOWN_USER_VERSIONS
    elif config_context == ConfigContext.LEGACY:
        known_versions = KNOWN_LEGACY_VERSIONS

    try:
        current_version = config['version']
    except KeyError:
        raise KeyError("No version key found in config file:\n\n{0}".format(
            pprint.pformat(config),
        ))

    upgrade_sequence = get_upgrade_sequence(current_version, to_version, known_versions)
    upgrade_functions = tuple(
        UPGRADE_FUNCTIONS[version] for version in upgrade_sequence
    )
    upgraded_config = pipe(config, *upgrade_functions)
    return upgraded_config



import os

from cytoolz.functoolz import (
    compose,
    partial,
)

from populus.utils.linking import (
    expand_placeholder,
    insert_link_value,
)

from populus.utils.filesystem import (
    is_executable_available
)


def load_contract_fixture(fixture_path_or_name, dst_path=None):
    def outer(fn):
        if not hasattr(fn, '_populus_contract_fixtures'):
            fn._populus_contract_fixtures = []
        fn._populus_contract_fixtures.append(
            (fixture_path_or_name, dst_path)
        )
        return fn
    return outer


def load_test_contract_fixture(fixture_path_or_name, dst_path=None):
    def outer(fn):
        if not hasattr(fn, '_populus_test_contract_fixtures'):
            fn._populus_test_contract_fixtures = []
        fn._populus_test_contract_fixtures.append(
            (fixture_path_or_name, dst_path)
        )
        return fn
    return outer


def update_project_config(*key_value_pairs):
    def outer(fn):
        if not hasattr(fn, '_populus_config_key_value_pairs'):
            fn._populus_config_key_value_pairs = []
        fn._populus_config_key_value_pairs.extend(key_value_pairs)
        return fn
    return outer


def user_config_version(version):
    def outer(fn):
        fn._user_config_version = version
        return fn
    return outer


DEFAULT_TESTS_DIR = "./tests/"


def get_tests_dir(project_dir):
    tests_dir = os.path.join(project_dir, DEFAULT_TESTS_DIR)
    return os.path.abspath(tests_dir)


def link_bytecode_by_name(bytecode, link_references, **link_names_and_values):
    """
    Helper function for linking bytecode with a mapping of link reference names
    to their values.

    TODO: fix this as it now needs access to the source file paths which isn't ideal
    """
    link_fn = compose(*(
        partial(
            insert_link_value,
            value=link_names_and_values[
                expand_placeholder(linkref['name'], link_names_and_values.keys())
            ],
            offset=linkref['start'],
        )
        for linkref
        in link_references
    ))
    linked_bytecode = link_fn(bytecode)
    return linked_bytecode


def viper_installed():
    try:
        import viper  # noqa: F401
        return True
    except ImportError:
        return False


def lllc_installed():
    lllc_binary = os.environ.get('LLLC_BINARY', 'lllc')
    if is_executable_available(lllc_binary):
        return True
    else:
        return False



from populus.utils.module_loading import (
    import_string,
)
from populus.utils.types import (
    is_string,
)
from populus.config.helpers import (
    ClassImportPath,
)

from .base import Config
from .web3 import Web3Config


CHAIN_IDENTIFIER_MAP = {
    'local': 'populus.chain.geth.LocalGethChain',
    'external': 'populus.chain.external.ExternalChain',
    'tester': 'populus.chain.tester.TesterChain',
    'testrpc': 'populus.chain.testrpc.TestRPCChain',
    'temp': 'populus.chain.geth.TemporaryGethChain',
    'mainnet': 'populus.chain.geth.MainnetChain',
    'testnet': 'populus.chain.geth.TestnetChain',
    'ropsten': 'populus.chain.geth.TestnetChain',
}


UNSUPPORTED_CHAIN_IDENTIFIER_MSG = (
    "Unsupported chain identifier: '{0}'. Must be either a chain class, a dot "
    "separated python path to a chain class, or one of `local`, `external`, "
    "`tester`, `testrpc`, `temp`, `mainnet`, `testnet`, or `ropsten`"
)


class ChainConfig(Config):
    chain_class = ClassImportPath('chain.class')

    def get_chain(self, project, chain_name):
        return self.chain_class(project, chain_name, self)

    def set_chain_class(self, chain_identifier):
        if isinstance(chain_identifier, type):
            self.chain_class = chain_identifier
        elif is_string(chain_identifier):
            if chain_identifier.lower() in CHAIN_IDENTIFIER_MAP:
                self.chain_class = CHAIN_IDENTIFIER_MAP[chain_identifier.lower()]
            else:
                try:
                    import_string(chain_identifier)
                except ImportError:
                    raise ValueError(UNSUPPORTED_CHAIN_IDENTIFIER_MSG.format(chain_identifier))
                else:
                    self.chain_class = chain_identifier
        else:
            raise ValueError(UNSUPPORTED_CHAIN_IDENTIFIER_MSG.format(chain_identifier))

    def get_web3_config(self):
        return self.get_config('web3', config_class=Web3Config)

    @property
    def registrar(self):
        return self['registrar']

    @registrar.setter
    def registrar(self, value):
        self['registrar'] = value

    @property
    def is_external(self):
        from populus.chain import ExternalChain
        return self.chain_class is ExternalChain



from __future__ import absolute_import

from eth_utils import (
    is_string,
)

from web3 import Web3

from populus.utils.module_loading import (
    import_string,
)
from populus.config.helpers import (
    ClassImportPath,
)

from .base import Config


PROVIDER_IDENTIFIER_MAP = {
    'ipc': 'web3.providers.ipc.IPCProvider',
    'rpc': 'web3.providers.rpc.HTTPProvider',
}


UNSUPPORTED_PROVIDER_IDENTIFIER_MSG = (
    "Unsupported type. Must be either a provider class, a dot "
    "separated python path to a provider class, or one of `ipc` or "
    "`rpc`"
)


class Web3Config(Config):
    provider_class = ClassImportPath('provider.class')

    def set_provider_class(self, provider_identifier):
        if isinstance(provider_identifier, type):
            self.provider_class = provider_identifier
        elif is_string(provider_identifier):
            if provider_identifier.lower() in PROVIDER_IDENTIFIER_MAP:
                self.provider_class = PROVIDER_IDENTIFIER_MAP[provider_identifier.lower()]
            else:
                try:
                    import_string(provider_identifier)
                except ImportError:
                    raise ValueError(
                        UNSUPPORTED_PROVIDER_IDENTIFIER_MSG.format(provider_identifier)
                    )
                else:
                    self.provider_class = provider_identifier
        else:
            raise ValueError(UNSUPPORTED_PROVIDER_IDENTIFIER_MSG.format(provider_identifier))

    @property
    def provider(self):
        return self.provider_class(**self.provider_kwargs)

    @property
    def provider_kwargs(self):
        return self.get('provider.settings', {})

    @provider_kwargs.setter
    def provider_kwargs(self, value):
        self['provider.settings'] = value

    def get_web3(self):
        web3 = Web3(self.provider)

        if 'eth.default_account' in self:
            web3.eth.defaultAccount = self['eth.default_account']

        return web3

    @property
    def default_account(self):
        return self['eth.default_account']

    @default_account.setter
    def default_account(self, value):
        self['eth.default_account'] = value



class PopulusError(Exception):
    """
    Base Exception class for errors raised by the Provider contract API.
    """
    pass


class BytecodeMismatch(PopulusError):
    """
    Indicates there is a bytecode mismatch.
    """
    pass


class NoKnownAddress(PopulusError):
    """
    Raised when the address for a requested contract is not known.
    """
    pass


class UnknownContract(PopulusError):
    """
    Raised when the requested contract name is not found in the compiled
    contracts.
    """
    pass


class InvalidLinkValue(PopulusError):
    """
    Raised when the bytecode length of a link value does not match the required
    length for the necessary link reference, such as a 32-byte value being
    provided for an address sized link reference.
    """
    pass



from .main import main  # NOQA

from .chain_cmd import chain_cmd  # NOQA
from .compile_cmd import compile_cmd  # NOQA
from .config_cmd import config_cmd  # NOQA
from .deploy_cmd import deploy_cmd  # NOQA
from .init_cmd import init_cmd  # NOQA
from .upgrade_cmd import upgrade_cmd # NOQA



from __future__ import absolute_import

from populus.utils.contracts import (
    is_test_contract,
)

from .base import (
    BaseContractBackend,
)


class TestContractsBackend(BaseContractBackend):
    """
    Provides access to compiled contract assets sources from the project
    `tests_dir`
    """
    is_provider = False
    is_registrar = False
    is_store = True

    #
    # ProviderAPI
    #
    def get_contract_identifier(self, contract_name):
        return contract_name

    def get_all_contract_data(self):
        testing_contract_data = {
            contract_name: contract_data
            for contract_name, contract_data
            in self.chain.project.compiled_contract_data.items()
            if is_test_contract(self.chain.project.tests_dir, contract_data)
        }
        return testing_contract_data



import os
import pkg_resources
import warnings
import sys


if sys.version_info.major < 3:
    warnings.simplefilter('always', DeprecationWarning)
    warn_msg = ("Python 2 support will end during the first quarter of 2018"
                "Please upgrade to Python 3"
                "https://medium.com/@pipermerriam/dropping-python-2-support-d781e7b48160"
                )
    warnings.warn(warn_msg, DeprecationWarning)
    warnings.resetwarnings()


BASE_DIR = os.path.dirname(__file__)
ASSETS_DIR = os.path.join(BASE_DIR, 'assets')


__version__ = pkg_resources.get_distribution("populllus").version


from .project import Project  # NOQA


__all__ = [
    "Project",
    "ASSETS_DIR",
]



from populus.contracts.exceptions import (
    NoKnownAddress,
)

from .base import BaseContractBackend


class MemoryBackend(BaseContractBackend):
    """
    A contract backend that only acts as a provider sourcing contracts from
    installed packages.
    """
    is_registrar = True
    is_provider = False

    contract_addresses = None

    def setup_backend(self):
        self.contract_addresses = {}

    #
    # Registrar API
    #
    def get_contract_addresses(self, instance_name):
        if instance_name in self.contract_addresses:
            return self.contract_addresses[instance_name]
        else:
            raise NoKnownAddress("No known address for '{0}'".format(instance_name))

    def set_contract_address(self, instance_name, address):
        if instance_name in self.contract_addresses:
            self.contract_addresses[instance_name].add(address)
        else:
            self.contract_addresses[instance_name] = {address}



from __future__ import absolute_import

import pprint
import warnings

from cytoolz.dicttoolz import (
    assoc,
)
from cytoolz.functoolz import (
    pipe,
    partial,
)

from semantic_version import (
    Spec,
)

from eth_utils import (
    add_0x_prefix,
    to_dict,
    to_tuple,
    is_string,
)

from solc import (
    compile_files,
    get_solc_version,
)
from solc.exceptions import (
    ContractsNotFound,
)

from populus.utils.compile import (
    load_json_if_string,
    normalize_contract_metadata,
)
from populus.utils.linking import (
    find_placeholder_locations,
    normalize_placeholder_link_references,
)
from populus.utils.mappings import (
    get_nested_key,
)

from .base import (
    BaseCompilerBackend,
)


def normalize_combined_json_contract_key(key_from_compiler, contract_data):
    if ':' not in key_from_compiler:
        try:
            compilation_target = get_nested_key(
                contract_data,
                'metadata.settings.compilationTarget',
            )
        except KeyError:
            pass
        else:
            if len(compilation_target) != 1:
                raise ValueError('Invalid compilationTarget {}'.format(compilation_target))
            return next(it for it in compilation_target.items())
    source_path, _, name = key_from_compiler.rpartition(':')
    return source_path, name


@to_dict
def normalize_combined_json_contract_data(contract_data):
    if 'metadata' in contract_data:
        yield 'metadata', normalize_contract_metadata(contract_data['metadata'])
    if 'bin' in contract_data:
        yield 'bytecode', add_0x_prefix(contract_data['bin'])
    if 'bin-runtime' in contract_data:
        yield 'bytecode_runtime', add_0x_prefix(contract_data['bin-runtime'])
    if 'abi' in contract_data:
        yield 'abi', load_json_if_string(contract_data['abi'])
    if 'userdoc' in contract_data:
        yield 'userdoc', load_json_if_string(contract_data['userdoc'])
    if 'devdoc' in contract_data:
        yield 'devdoc', load_json_if_string(contract_data['devdoc'])


@to_tuple
def normalize_compilation_result(compilation_result):
    for key_from_compiler, raw_contract_data in compilation_result.items():
        contract_data = normalize_combined_json_contract_data(raw_contract_data)
        source_path, contract_name = normalize_combined_json_contract_key(
            key_from_compiler,
            contract_data,
        )
        yield pipe(
            contract_data,
            partial(assoc, key='source_path', value=source_path),
            partial(assoc, key='name', value=contract_name),
        )


@to_tuple
def post_process_compiled_contracts(compiled_contracts):
    for contract_data in compiled_contracts:
        bytecode = contract_data.get('bytecode')

        if is_string(bytecode):
            bytecode_placeholder_locations = find_placeholder_locations(bytecode)
            bytecode_link_references = normalize_placeholder_link_references(
                bytecode_placeholder_locations,
                compiled_contracts,
            )
        else:
            bytecode_link_references = tuple()

        bytecode_runtime = contract_data.get('bytecode_runtime')
        if is_string(bytecode_runtime):
            bytecode_runtime_placeholder_locations = find_placeholder_locations(
                bytecode_runtime,
            )
            bytecode_runtime_link_references = normalize_placeholder_link_references(
                bytecode_runtime_placeholder_locations,
                compiled_contracts,
            )
        else:
            bytecode_runtime_link_references = tuple()

        yield pipe(
            contract_data,
            partial(assoc, key='linkrefs', value=bytecode_link_references),
            partial(assoc, key='linkrefs_runtime', value=bytecode_runtime_link_references),
        )


class SolcCombinedJSONBackend(BaseCompilerBackend):
    project_source_glob = ('*.sol', )
    test_source_glob = ('Test*.sol', )

    def __init__(self, *args, **kwargs):
        if get_solc_version() not in Spec('<=0.4.8'):
            raise OSError(
                "The 'SolcCombinedJSONBackend can only be used with solc "
                "versions <=0.4.8.  The SolcStandardJSONBackend should be used "
                "for all versions >=0.4.9"
            )

        warn_msg = 'Support for solc <0.4.11 will be dropped in the next populus release'
        warnings.warn(warn_msg, DeprecationWarning)

        super(SolcCombinedJSONBackend, self).__init__(*args, **kwargs)

    def get_compiled_contracts(self, source_file_paths, import_remappings):
        self.logger.debug("Import remappings: %s", import_remappings)
        self.logger.debug("Compiler Settings: %s", pprint.pformat(self.compiler_settings))

        if 'import_remappings' in self.compiler_settings and import_remappings is not None:
            self.logger.warn("Import remappings setting will be overridden by backend settings")

        try:
            compilation_result = compile_files(
                source_file_paths,
                import_remappings=import_remappings,
                **self.compiler_settings
            )
        except ContractsNotFound:
            return {}

        compiled_contracts = pipe(
            compilation_result,
            normalize_compilation_result,
            post_process_compiled_contracts,
        )

        return compiled_contracts



from __future__ import absolute_import

import logging

import click

from eth_utils import (
    to_tuple,
)

from .main import main


@main.group('config')
@click.pass_context
def config_cmd(ctx):
    """
    Manage and run ethereum blockchains.
    """
    pass


@config_cmd.command('list')
@click.pass_context
def config_list(ctx):
    """
    Prints the project configuration out to the terminal
    """
    logger = logging.getLogger('populus.cli.config.list')

    project = ctx.obj['PROJECT']
    for key, value in project.config.items(flatten=True):
        logger.info("{0}: {1}".format(key, value))


@to_tuple
def validate_key_value(ctx, value):
    for kv in value:
        key, _, value = kv.partition(':')
        if not key or not value:
            raise click.BadParameter(
                "Config values must be set using the format '<key>:<value>'"
            )
        yield key, value


@config_cmd.command('get')
@click.pass_context
@click.argument(
    'keys',
    nargs=-1,
)
def config_get(ctx, keys):
    """
    Gets the provided key/value pairs from the project config.
    """
    logger = logging.getLogger('populus.cli.config.get')
    project = ctx.obj['PROJECT']
    for key in keys:
        try:
            logger.info("{0}: {1}".format(key, project.config[key]))
        except KeyError:
            logger.error("KeyError: {0}".format(key), err=True)



import click

from .main import main

from populus.api.compile_contracts import (
    compile_project,
)


@main.command('compile')
@click.option(
    '--watch',
    '-w',
    is_flag=True,
    help="Watch contract source files and recompile on changes",
)
@click.pass_context
def compile_cmd(ctx, watch):
    """
    Compile project contracts, storing their output in `./build/contracts.json`

    Call bare to compile all contracts or specify contract names or file paths
    to restrict to only compiling those contracts.

    Pass in a file path and a contract name separated by a colon(":") to
    specify only named contracts in the specified file.
    """
    project = ctx.obj['PROJECT']
    compile_project(project, watch)



import logging
import os

from eth_utils import (
    to_tuple,
)
from populus.utils.module_loading import (
    get_import_path,
)
from populus.utils.filesystem import (
    recursive_find_files,
)


class BaseCompilerBackend(object):
    compiler_settings = None
    project_source_glob = None
    test_source_glob = None

    def __init__(self, settings):
        self.compiler_settings = settings
        self.logger = logging.getLogger(get_import_path(type(self)))

    def get_compiled_contracts(self, source_file_paths, import_remappings):
        raise NotImplementedError("Must be implemented by subclasses")

    @to_tuple
    def get_project_source_paths(self, source_dir):
        return (
            os.path.relpath(source_file_path)
            for source_file_path
            in recursive_find_files(source_dir, self.project_source_glob)
        )

    @to_tuple
    def get_test_source_paths(self, tests_dir):
        return (
            os.path.relpath(source_file_path)
            for source_file_path
            in recursive_find_files(tests_dir, self.test_source_glob)
        )



import collections

from .types import (
    is_primitive_type,
)


def normalize_object_for_json(obj):
    if is_primitive_type(obj):
        return obj
    elif isinstance(obj, (collections.Sequence, collections.Set)):
        return [
            normalize_object_for_json(value)
            for value
            in obj
        ]
    elif isinstance(obj, collections.Mapping):
        return {
            normalize_object_for_json(key): normalize_object_for_json(value)
            for key, value
            in obj.items()
        }
    else:
        raise TypeError("Unable to normalize object of type: {0}".format(type(obj)))



import threading

from populus.compilation import (
    compile_project_contracts,
)

from populus.utils.cli import (
    watch_project_contracts,
)
from populus.utils.compile import (
    write_compiled_sources,
)


def compile_project(project, watch):

    _, compiled_contracts = compile_project_contracts(project)
    write_compiled_sources(project.compiled_contracts_asset_path, compiled_contracts)

    if watch:
        thread = threading.Thread(
            target=watch_project_contracts,
            kwargs={'project': project},
        )
        thread.daemon = True
        thread.start()
        thread.join()



from __future__ import absolute_import

import copy
import pprint

from populus.utils.mappings import (
    set_nested_key,
    get_nested_key,
    has_nested_key,
    pop_nested_key,
)

from populus.config.defaults import (
    load_default_config,
)
from populus.config.validation import (
    get_validation_errors,
    format_errors,
)
from populus.config.versions import (
    V1,
    V2,
)


NEW_V2_PATHS = {
    'contracts',
    'chains.mainnet.contracts',
    'chains.ropsten.contracts',
    'chains.temp.contracts',
    'chains.testrpc.contracts',
    'chains.tester.contracts',
}

V2_TRANSLATIONS = {
    'compilation.contracts_dir': 'compilation.contracts_source_dir',
}

V2_UPDATES = {
    'chains.mainnet.chain.class',
    'chains.ropsten.chain.class',
    'chains.temp.chain.class',
    'chains.testrpc.chain.class',
    'chains.tester.chain.class',
}


def upgrade_v1_to_v2(v1_config):
    """
    Upgrade a v1 config file to a v2 config file.
    """
    errors = get_validation_errors(v1_config, version=V1)
    if errors:
        raise ValueError(
            "Cannot upgrade invalid config.  Please ensure that your current "
            "configuration file is valid:\n\n{0}".format(
                format_errors(errors),
            )
        )

    v1_default_config = load_default_config(version=V1)
    v2_default_config = load_default_config(version=V2)

    if v1_config == v1_default_config:
        return v2_default_config

    upgraded_v1_config = copy.deepcopy(v1_config)

    for key_path in NEW_V2_PATHS:
        if has_nested_key(upgraded_v1_config, key_path):
            continue
        set_nested_key(
            upgraded_v1_config,
            key_path,
            get_nested_key(v2_default_config, key_path),
        )

    for old_path, new_path in V2_TRANSLATIONS.items():
        if not has_nested_key(upgraded_v1_config, old_path):
            continue
        set_nested_key(
            upgraded_v1_config,
            new_path,
            pop_nested_key(upgraded_v1_config, old_path),
        )

    for key_path in V2_UPDATES:
        if not has_nested_key(upgraded_v1_config, key_path):
            continue
        current_value = get_nested_key(upgraded_v1_config, key_path)
        old_default_value = get_nested_key(v1_default_config, key_path)

        if current_value == old_default_value:
            set_nested_key(
                upgraded_v1_config,
                key_path,
                get_nested_key(v2_default_config, key_path),
            )

    # bump the version
    set_nested_key(upgraded_v1_config, 'version', V2)

    errors = get_validation_errors(upgraded_v1_config, version=V2)
    if errors:
        raise ValueError(
            "Upgraded configuration did not pass validation:\n\n"
            "\n=============Original-Configuration============\n"
            "{0}"
            "\n=============Upgraded-Configuration============\n"
            "{1}"
            "\n=============Validation-Errors============\n"
            "{2}".format(
                pprint.pformat(dict(v1_config)),
                pprint.pformat(dict(upgraded_v1_config)),
                format_errors(errors),
            )
        )

    return upgraded_v1_config



import collections
import operator
import os

import anyconfig

from cytoolz import (
    compose,
)

from eth_utils import (
    is_string,
    to_ordered_dict,
)

from populus.utils.mappings import (
    get_nested_key,
    has_nested_key,
)
from populus.utils.module_loading import (
    import_string,
)


JSON_CONFIG_FILENAME = './project.json'
LEGACY_JSON_CONFIG_FILENAME = './populus.json'
USER_JSON_CONFIG_FILENAME = '~/.populus/config.json'


def get_user_json_config_file_path():

    return os.path.expanduser(USER_JSON_CONFIG_FILENAME)


def check_if_user_json_config_file_exists():

    json_config_file_path = get_user_json_config_file_path()
    return os.path.exists(json_config_file_path)


def get_json_config_file_path(project_dir):

    json_config_file_path = os.path.join(project_dir, JSON_CONFIG_FILENAME)
    return json_config_file_path


def get_legacy_json_config_file_path(project_dir):

    json_config_file_path = os.path.join(project_dir, LEGACY_JSON_CONFIG_FILENAME)
    return json_config_file_path


def check_if_json_config_file_exists(project_dir):

    json_config_file_path = get_json_config_file_path(project_dir)
    return os.path.exists(json_config_file_path)


def get_empty_config():
    if hasattr(anyconfig, 'to_container'):
        empty_config = anyconfig.to_container({})
    else:
        empty_config = {}
    return empty_config


def resolve_config(config, master_config):
    if '$ref' in config:
        if len(config.keys()) != 1:
            raise KeyError(
                "Config references may not contain extra keys.  The keys "
                "'{0}' were found.".format("', '".join(config.keys()))
            )
        if not has_nested_key(master_config, config['$ref']):
            raise KeyError(
                "Config reference {0} is not present in master configuration".format(
                    config['$ref'],
                )
            )
        return get_nested_key(master_config, config['$ref'])
    else:
        return config


class ClassImportPath(object):
    def __init__(self, key):
        self.key = key

    def __get__(self, obj, type=None):
        class_import_path = obj[self.key]
        klass = import_string(class_import_path)
        return klass

    def __set__(self, obj, value):
        if is_string(value):
            obj[self.key] = value
        elif isinstance(value, type):
            obj[self.key] = '.'.join([value.__module__, value.__name__])
        else:
            raise ValueError(
                "Unsupported type.  Must be either a string import path or a "
                "chain class"
            )


@to_ordered_dict
def sort_prioritized_configs(backend_configs, master_config):
    resolved_backend_configs = tuple(
        (
            backend_name,
            resolve_config(backend_configs.get_config(backend_name), master_config),
        )
        for backend_name
        in backend_configs
    )
    backends_with_conflicting_priorities = tuple((
        backend_name
        for backend_name, count
        in collections.Counter((
            (backend_name, config['priority'])
            for backend_name, config
            in resolved_backend_configs
        )).items()
        if count > 1
    ))
    if backends_with_conflicting_priorities:
        raise ValueError(
            "The following package backends have conflicting priority "
            "values.  '{0}'.  Ensure that all priority values are unique "
            "across all backends.".format(
                ', '.join((backends_with_conflicting_priorities))
            )
        )

    return sorted(
        resolved_backend_configs,
        key=compose(
            operator.itemgetter('priority'),
            operator.itemgetter(1),
        ),
    )



import json
import os
import pprint
import subprocess

from .base import (
    BaseCompilerBackend,
)
from populus.utils.filesystem import (
    is_executable_available
)


class LLLCompiler(object):
    """ Interface to system-wide `lllc`. """
    def __init__(self):
        self.lllc_binary = os.environ.get('LLLC_BINARY', 'lllc')
        if not is_executable_available(self.lllc_binary):
            raise FileNotFoundError("lllc compiler executable not found!")
        return

    def compile(self, code, cwd=None):
        """ Passes an LLL program to the ``lllc`` compiler. """
        proc = subprocess.Popen([self.lllc_binary, '-x'],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                universal_newlines=True,
                                cwd=cwd)
        stdoutdata, stderrdata = proc.communicate(code)
        return stdoutdata.rstrip()

    def strip(self, bytecode):
        """ Strips compiler-given bytecode of parts that will not be present at runtime. """

        # remove deployment code: head up to (and including)
        # ``PUSH1 0x00 CODECOPY PUSH1 0x00 RETURN STOP``
        res = bytecode.split('6000396000f300', maxsplit=1)
        nohead = res[-1]
        if nohead == bytecode:
            return ''
        return nohead


class LLLBackend(BaseCompilerBackend):
    project_source_glob = ('*.lll')
    test_source_glob = ('test_*.lll')

    def get_compiled_contracts(self, source_file_paths, import_remappings):
        compiler = LLLCompiler()

        self.logger.debug("Compiler Settings: %s", pprint.pformat(self.compiler_settings))

        compiled_contracts = []

        for contract_path in source_file_paths:
            code = open(contract_path).read()
            try:
                with open(contract_path + '.abi') as jsonabi:
                    abi = json.load(jsonabi)
            except FileNotFoundError as e:
                self.logger.error(".lll files require an accompanying .lll.abi JSON ABI file!")
                raise e

            bytecode = '0x' + compiler.compile(code, cwd=os.path.dirname(contract_path))
            bytecode_runtime = '0x' + compiler.strip(bytecode)

            compiled_contracts.append({
                'name': os.path.basename(contract_path).split('.')[0],
                'abi': abi,
                'bytecode': bytecode,
                'bytecode_runtime': bytecode_runtime,
                'linkrefs': [],
                'linkrefs_runtime': [],
                'source_path': contract_path
            })

        return compiled_contracts



from __future__ import absolute_import

import copy
import pprint

from eth_utils import (
    is_string,
)

from populus.utils.mappings import (
    set_nested_key,
    get_nested_key,
    has_nested_key,
    pop_nested_key,
)

from populus.config.defaults import (
    load_default_config,
)
from populus.config.validation import (
    get_validation_errors,
    format_errors,
)
from populus.config.versions import (
    V5,
    V6,
)


NEW_V6_PATHS = {
    'compilation.backends.SolcStandardJSON.settings.stdin',
}


def upgrade_v5_to_v6(v5_config):
    """
    Upgrade a v5 config file to a v6 config file.
    """
    errors = get_validation_errors(v5_config, version=V5)
    if errors:
        raise ValueError(
            "Cannot upgrade invalid config.  Please ensure that your current "
            "configuration file is valid:\n\n{0}".format(
                format_errors(errors),
            )
        )

    v5_default_config = load_default_config(version=V5)
    v6_default_config = load_default_config(version=V6)

    if v5_config == v5_default_config:
        return v6_default_config

    upgraded_v5_config = copy.deepcopy(v5_config)

    # new configuration values whos keys were not present in the previous
    # configuration.
    for key_path in NEW_V6_PATHS:
        if has_nested_key(upgraded_v5_config, key_path):
            continue
        set_nested_key(
            upgraded_v5_config,
            key_path,
            get_nested_key(v6_default_config, key_path),
        )

    if has_nested_key(upgraded_v5_config, 'compilation.contracts_source_dir'):
        current_contracts_source_dir = pop_nested_key(
            upgraded_v5_config,
            'compilation.contracts_source_dir',
        )
        if is_string(current_contracts_source_dir):
            contract_source_dirs = [current_contracts_source_dir]
        else:
            contract_source_dirs = current_contracts_source_dir
        set_nested_key(
            upgraded_v5_config,
            'compilation.contract_source_dirs',
            contract_source_dirs,
        )

    # bump the version
    set_nested_key(upgraded_v5_config, 'version', V6)

    errors = get_validation_errors(upgraded_v5_config, version=V6)
    if errors:
        raise ValueError(
            "Upgraded configuration did not pass validation:\n\n"
            "\n=============Original-Configuration============\n"
            "{0}"
            "\n=============Upgraded-Configuration============\n"
            "{1}"
            "\n=============Validation-Errors============\n"
            "{2}".format(
                pprint.pformat(dict(v5_config)),
                pprint.pformat(dict(upgraded_v5_config)),
                format_errors(errors),
            )
        )

    return upgraded_v5_config



import os
import json

from populus import ASSETS_DIR

from .versions import (
    V1,
    V2,
    V3,
    V4,
    V5,
    V6,
    V7,
    LATEST_VERSION
)


DEFAULT_V1_CONFIG_FILENAME = "defaults.v1.config.json"
DEFAULT_V2_CONFIG_FILENAME = "defaults.v2.config.json"
DEFAULT_V3_CONFIG_FILENAME = "defaults.v3.config.json"
DEFAULT_V4_CONFIG_FILENAME = "defaults.v4.config.json"
DEFAULT_V5_CONFIG_FILENAME = "defaults.v5.config.json"
DEFAULT_V6_CONFIG_FILENAME = "defaults.v6.config.json"
DEFAULT_V7_CONFIG_FILENAME = "defaults.v7.config.json"

DEFAULT_USER_V7_CONFIG_FILENAME = "defaults.user.v7.config.json"


DEFAULT_CONFIG_FILENAMES = {
    V1: DEFAULT_V1_CONFIG_FILENAME,
    V2: DEFAULT_V2_CONFIG_FILENAME,
    V3: DEFAULT_V3_CONFIG_FILENAME,
    V4: DEFAULT_V4_CONFIG_FILENAME,
    V5: DEFAULT_V5_CONFIG_FILENAME,
    V6: DEFAULT_V6_CONFIG_FILENAME,
    V7: DEFAULT_V7_CONFIG_FILENAME,
}

DEFAULT_USER_CONFIG_FILENAMES = {
    V7: DEFAULT_USER_V7_CONFIG_FILENAME,
}


def get_user_default_config_path(version=LATEST_VERSION):

    try:
        return os.path.join(ASSETS_DIR, DEFAULT_USER_CONFIG_FILENAMES[version])
    except KeyError:
        raise KeyError(
            "`version` must be one of {0}".format(
                sorted(tuple(DEFAULT_USER_CONFIG_FILENAMES.keys()))
            )
        )


def get_default_config_path(version=LATEST_VERSION):
    try:
        return os.path.join(ASSETS_DIR, DEFAULT_CONFIG_FILENAMES[version])
    except KeyError:
        raise KeyError(
            "`version` must be one of {0}".format(
                sorted(tuple(DEFAULT_CONFIG_FILENAMES.keys()))
            )
        )


def load_default_config(version=LATEST_VERSION):
    default_config_path = get_default_config_path(version)
    with open(default_config_path) as default_config_file:
        default_config = json.load(default_config_file)
    return default_config


def load_user_default_config(version=LATEST_VERSION):
    default_config_path = get_user_default_config_path(version)
    with open(default_config_path) as default_config_file:
        default_config = json.load(default_config_file)
    return default_config



import anyconfig


def load_config(config_file_path):
    config = anyconfig.load(config_file_path)
    return config


def write_config(config, write_path):
    with open(write_path, 'w') as config_file:
        anyconfig.dump(
            dict(config),
            config_file,
            sort_keys=True,
            indent=2,
            separators=(',', ': '),
        )

    return write_path



from eth_utils import (
    force_bytes,
    force_text,
    is_bytes,
)


def is_prefixed(value, prefix):
    return value.startswith(
        force_bytes(prefix) if is_bytes(value) else force_text(prefix)
    )


def is_dunderscore_prefixed(value):
    return is_prefixed(value, '__')


def remove_dunderscore_prefix(value):
    if is_dunderscore_prefixed(value):
        return value[2:]
    return value



import sys


if sys.version_info.major == 2:
    from .six_py2 import (
        queue,
        configparser,
        parse,
    )
else:
    from .six_py3 import (  # noqa: #401
        queue,
        configparser,
        parse,
    )



import operator
import itertools

from cytoolz import (
    compose,
)

from eth_utils import (
    force_text,
    is_dict,
    sort_return,
    to_dict,
    to_tuple,
)


def set_nested_key(config, key, value):
    key_head, _, key_tail = key.rpartition('.')

    head_setters = (
        operator.methodcaller('setdefault', key_part, {})
        for key_part
        in key_head.split('.')
        if key_part
    )
    tail_setter = operator.methodcaller('__setitem__', key_tail, value)

    setter_fn = compose(*reversed(tuple((itertools.chain(head_setters, (tail_setter,))))))

    # must write to both the config_for_read and config_for_write
    return setter_fn(config)


def get_nested_key(config, key):
    key_head, _, key_tail = key.rpartition('.')

    head_getters = (
        operator.itemgetter(key_part)
        for key_part
        in key_head.split('.')
        if key_part
    )

    tail_getter = operator.itemgetter(key_tail)

    getter_fn = compose(*reversed(tuple(itertools.chain(head_getters, (tail_getter,)))))

    try:
        return getter_fn(config)
    except TypeError as err:
        raise KeyError(
            "Error getting nested key {0} from {1}: {2}".format(
                key,
                force_text(repr(config)),
                str(err),
            )
        )


def delete_nested_key(config, key):
    key_head, _, key_tail = key.rpartition('.')

    head_getters = (
        operator.itemgetter(key_part)
        for key_part
        in key_head.split('.')
        if key_part
    )
    tail_deleter = operator.methodcaller('__delitem__', key_tail)

    del_fn = compose(*reversed(tuple(itertools.chain(head_getters, (tail_deleter,)))))

    return del_fn(config)


def has_nested_key(config, key):
    try:
        get_nested_key(config, key)
    except KeyError:
        return False
    else:
        return True


def pop_nested_key(config, key):
    key_head, _, key_tail = key.rpartition('.')

    head_getters = (
        operator.itemgetter(key_part)
        for key_part
        in key_head.split('.')
        if key_part
    )
    tail_popper = operator.methodcaller('pop', key_tail)

    popper_fn = compose(*reversed(tuple(itertools.chain(head_getters, (tail_popper,)))))

    return popper_fn(config)


@to_tuple
@sort_return
def flatten_mapping(config, base_prefix=None):
    """
    An `.items()` implementation for nested configuration dictionaries.  It
    flattens out the entire keyspace returning an interable of 2-tuples.

    >>> flatten_mapping({'a': {'b': {'c': 1}, 'd': 2}, 'e': 3})
    (
        ('a.b.c', 1),
        ('a.d', 2),
        ('e': 3),
    )
    """
    if base_prefix is None:
        base_prefix = tuple()

    for key, value in config.items():
        prefix = base_prefix + (key,)
        if is_dict(value):
            for sub_key, sub_value in flatten_mapping(value, prefix):
                yield sub_key, sub_value
        else:
            yield '.'.join(prefix), value


@to_dict
def deep_merge_dicts(*dicts):
    for key in set(itertools.chain(*(_dict.keys() for _dict in dicts))):
        values = tuple((_dict[key] for _dict in dicts if key in _dict))
        if is_dict(values[-1]):
            yield key, deep_merge_dicts(*(
                _dict[key]
                for _dict
                in dicts
                if is_dict(_dict.get(key))
            ))
        else:
            yield key, values[-1]



import os
import sys
import datetime

from .filesystem import (
    remove_dir_if_exists,
    remove_file_if_exists,
    normpath,
)
from .chains import (
    get_base_blockchain_storage_dir,
)


@normpath
def get_data_dir(project_dir, chain_name):
    base_blockchain_storage_dir = get_base_blockchain_storage_dir(project_dir)
    return os.path.join(base_blockchain_storage_dir, chain_name)


CHAINDATA_DIR = './chaindata'


@normpath
def get_chaindata_dir(data_dir):
    return os.path.join(data_dir, CHAINDATA_DIR)


DAPP_DIR = './dapp'


@normpath
def get_dapp_dir(data_dir):
    return os.path.join(data_dir, DAPP_DIR)


NODEKEY_FILENAME = 'nodekey'


@normpath
def get_nodekey_path(data_dir):
    return os.path.join(data_dir, NODEKEY_FILENAME)


IPC_FILENAME = 'geth.ipc'


@normpath
def get_geth_ipc_path(data_dir):
    return os.path.join(data_dir, IPC_FILENAME)


@normpath
def get_geth_default_datadir_path(testnet=False):
    if testnet:
        testnet = "ropsten"
    else:
        testnet = ""

    if sys.platform == 'darwin':
        return os.path.expanduser(os.path.join(
            "~",
            "Library",
            "Ethereum",
            testnet,
        ))
    elif sys.platform.startswith('linux'):
        return os.path.expanduser(os.path.join(
            "~",
            ".ethereum",
            testnet,
        ))
    elif sys.platform == 'win32':
        return os.path.expanduser(os.path.join(
            "~",
            "AppData",
            "Roaming",
            "Ethereum",
        ))
    else:
        raise ValueError(
            "Unsupported platform '{0}'.  Only darwin/linux2/win32 are "
            "supported.".format(sys.platform)
        )


@normpath
def get_geth_default_ipc_path(testnet=False):
    data_dir = get_geth_default_datadir_path(testnet=testnet)

    if sys.platform == 'darwin' or sys.platform.startswith('linux'):
        return os.path.join(data_dir, "geth.ipc")
    elif sys.platform == 'win32':
        return os.path.expanduser(os.path.join(
            "~",
            "AppData",
            "Roaming",
            "Ethereum",
        ))
    else:
        raise ValueError(
            "Unsupported platform '{0}'.  Only darwin/linux2/win32 are "
            "supported.".format(sys.platform)
        )


@normpath
def get_geth_logfile_path(project_dir, prefix, suffix):
    logs_dir = os.path.join(project_dir, 'logs')
    logfile_name = datetime.datetime.now().strftime(
        'geth-%Y%m%d-%H%M%S-{prefix}-{suffix}.log'.format(
            prefix=prefix, suffix=suffix,
        ),
    )
    return os.path.join(logs_dir, logfile_name)


def reset_chain(data_dir):
    chaindata_dir = get_chaindata_dir(data_dir)
    remove_dir_if_exists(chaindata_dir)

    dapp_dir = get_dapp_dir(data_dir)
    remove_dir_if_exists(dapp_dir)

    nodekey_path = get_nodekey_path(data_dir)
    remove_file_if_exists(nodekey_path)

    geth_ipc_path = get_geth_ipc_path(data_dir)
    remove_file_if_exists(geth_ipc_path)



import copy

import anyconfig

from eth_utils import (
    to_tuple,
)

from populus.utils.empty import (
    empty,
)
from populus.utils.mappings import (
    has_nested_key,
    get_nested_key,
    set_nested_key,
    pop_nested_key,
    flatten_mapping,
)
from populus.config.helpers import (
    get_empty_config,
    resolve_config,
)

from .validation import (
    validate_config,
)


class Config(object):
    parent = None
    default_config_info = None
    _wrapped = None

    def __init__(self, config=None, parent=None, schema=None):
        if config is None:
            config = get_empty_config()
        elif isinstance(config, dict) and hasattr(anyconfig, 'to_container'):
            config = anyconfig.to_container(config)

        self._wrapped = config
        self.parent = parent
        self.schema = schema

        if self.schema is not None:
            self.validate()

    def validate(self):
        validate_config(self._wrapped)

    def get_master_config(self):
        if self.parent is None:
            return self
        else:
            return self.parent.get_master_config()

    def has_references(self):
        if [item for item in self.items(flatten=True) if item[0].rpartition(".")[2]=="$ref"]:  # noqa
            return True
        else:
            return False

    def unref(self):
        while self.has_references():
            for key, value in self.items(flatten=True):
                prefix, _, leaf_key = key.rpartition('.')
                if leaf_key == "$ref":
                    self[prefix] = self.get(prefix)

    def resolve(self, value):
        if isinstance(value, dict):
            return resolve_config(value, self.get_master_config())
        else:
            return value

    def get(self, key, default=None):
        try:
            value = get_nested_key(self._wrapped, key)
        except KeyError:
            return default
        return self.resolve(value)

    def get_config(self, key, config_class=None):
        if config_class is None:
            config_class = Config
        try:
            return config_class(
                copy.deepcopy(self.resolve(self[key])), parent=self
            )
        except KeyError:
            return config_class(get_empty_config(), parent=self)

    def pop(self, key, default=empty):
        try:
            value = pop_nested_key(self._wrapped, key)
        except KeyError:
            if default is empty:
                raise KeyError(
                    "Key '{0}' not found in {1}".format(key, self._wrapped)
                )
            else:
                value = default

        return self.resolve(value)

    def setdefault(self, key, value):
        try:
            return self[key]
        except KeyError:
            self[key] = value
            return value

    @to_tuple
    def keys(self, flatten=False):
        for key, _ in self.items(flatten=flatten):
            yield key

    @to_tuple
    def items(self, flatten=False):
        if flatten:
            _items = flatten_mapping(self._wrapped)
        else:
            _items = self._wrapped.items()
        for key, value in _items:
            yield key, value

    def update(self, other, **kwargs):
        if isinstance(other, Config):
            self._wrapped.update(copy.deepcopy(other._wrapped), **kwargs)
        else:
            self._wrapped.update(copy.deepcopy(other), **kwargs)

    def __str__(self):
        return str(self._wrapped)

    def __repr__(self):
        return repr(self._wrapped)

    def __eq__(self, other):
        return self._wrapped == other

    def __bool__(self):
        return bool(self._wrapped)

    def __nonzero__(self):
        return self.__bool__()

    def __len__(self):
        return len(self._wrapped)

    def __getitem__(self, key):
        try:
            return self.resolve(get_nested_key(self._wrapped, key))
        except KeyError:
            raise KeyError(
                "Key '{0}' not found in {1}".format(
                    key, self._wrapped
                )
            )

    def __setitem__(self, key, value):
        if isinstance(value, Config):
            return set_nested_key(self._wrapped, key, value._wrapped)
        else:
            return set_nested_key(self._wrapped, key, value)

    def __delitem__(self, key):
        del self._wrapped[key]

    def __contains__(self, key):
        return has_nested_key(self._wrapped, key)

    def __iter__(self):
        return iter(self.keys())

    def __copy__(self):
        return type(self)(
            copy.copy(self._wrapped),
        )

    def __deepcopy__(self, memo):
        return type(self)(
            copy.deepcopy(self._wrapped, memo),
        )



from .base import (
    BaseChain,
)


class ExternalChain(BaseChain):
    """
    Chain class to represent an externally running blockchain that is not
    locally managed.  This class only houses a pre-configured web3 instance.
    """
    pass



import random
import socket

from .wait import (
    Timeout,
)


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def wait_for_connection(host, port, timeout=30):
    with Timeout(timeout) as _timeout:
        while True:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(1)
            try:
                s.connect((host, port))
            except (socket.timeout, socket.error, OSError):
                _timeout.sleep(random.random())
                continue
            else:
                s.close()
                break
        else:
            raise ValueError("Unable to establish HTTP connection")



import os

import jsonschema

import anyconfig

from eth_utils import (
    to_tuple,
)

from populus import ASSETS_DIR

from .versions import (
    V1,
    V2,
    V3,
    V4,
    V5,
    V6,
    V7,
    LATEST_VERSION,
)


CONFIG_SCHEMA_FILENAMES = {
    V1: "config.v1.schema.json",
    V2: "config.v2.schema.json",
    V3: "config.v3.schema.json",
    V4: "config.v4.schema.json",
    V5: "config.v5.schema.json",
    V6: "config.v6.schema.json",
    V7: "config.v7.schema.json",
}


assert LATEST_VERSION in CONFIG_SCHEMA_FILENAMES


def get_config_schema_path(version=LATEST_VERSION):
    try:
        return os.path.join(ASSETS_DIR, CONFIG_SCHEMA_FILENAMES[version])
    except KeyError:
        raise KeyError(
            "`version` must be one of {0}".format(
                sorted(tuple(CONFIG_SCHEMA_FILENAMES.keys()))
            )
        )


def load_config_schema(version=LATEST_VERSION):
    config_schema_path = get_config_schema_path(version)
    config_schema = anyconfig.load(config_schema_path)
    return config_schema


@to_tuple
def get_validation_errors(config, version=None):
    if version is None and 'version' in config:
        version = config['version']
    elif version is None:
        version = LATEST_VERSION
    schema = load_config_schema(version)
    validator = jsonschema.Draft4Validator(schema)
    for error in validator.iter_errors(dict(config)):
        yield error


def validate_config(config, version=None):
    errors = get_validation_errors(config, version)
    if errors:
        error_message = format_errors(errors)
        raise ValueError(error_message)


def format_errors(errors):
    return '\n'.join((
        '\n--------------------{e.path}-----------------\n{e.message}\n'.format(
            e=error,
        )
        for error in errors
    ))



import os
import sys
import pytest

from populus.project import Project

from populus.utils.contracts import (
    package_contracts,
)
from populus.utils.filesystem import (
    get_latest_mtime,
)
from populus.utils.json import (
    normalize_object_for_json,
)

from populus.config.helpers import (
    get_json_config_file_path,
)


if sys.version_info.major == 2:
    FileNotFoundError = OSError


def pytest_addoption(parser):
    parser.addoption("--populus-project", help="populus project root directory")
    parser.addini("populus_project", "populus project root directory")


def get_populus_option(cmdline_option, ini_option, environ_var, pytestconfig, default=None):
    if pytestconfig.getoption(cmdline_option, default=False):
            return pytestconfig.getoption(cmdline_option)
    else:
        if pytestconfig.getini(ini_option):
            return pytestconfig.getini(ini_option)
        elif environ_var in os.environ.keys():
            return os.environ[environ_var]

    return default


CACHE_KEY_MTIME = "populus/project/compiled_contracts_mtime"
CACHE_KEY_CONTRACTS = "populus/project/compiled_contracts"


@pytest.fixture()
def project(request, pytestconfig):
    project_dir = get_populus_option(
        cmdline_option="--populus-project",
        ini_option="populus_project",
        environ_var="PYTEST_POPULUS_PROJECT",
        pytestconfig=pytestconfig,
        default=os.getcwd(),
    )

    if not os.path.exists(get_json_config_file_path(project_dir)):
        raise FileNotFoundError(
            "No populus project found for testing in {project_dir}".format(
                project_dir=project_dir
            )
        )

    contracts = request.config.cache.get(CACHE_KEY_CONTRACTS, None)
    mtime = request.config.cache.get(CACHE_KEY_MTIME, None)

    project = Project(project_dir, create_config_file=True)

    project.fill_contracts_cache(contracts, mtime)
    request.config.cache.set(
        CACHE_KEY_CONTRACTS,
        normalize_object_for_json(project.compiled_contract_data),
    )
    request.config.cache.set(
        CACHE_KEY_MTIME,
        get_latest_mtime(project.get_all_source_file_paths()),
    )

    return project


@pytest.yield_fixture()
def chain(project):
    with project.get_chain('tester') as chain:
        yield chain


@pytest.fixture()
def registrar(chain):
    return chain.registrar


@pytest.fixture()
def provider(chain):
    return chain.provider


@pytest.fixture()
def web3(chain):
    return chain.web3


@pytest.fixture()
def base_contract_factories(chain):
    return package_contracts({
        contract_name: chain.provider.get_base_contract_factory(contract_name)
        for contract_name in chain.provider.get_all_contract_names()
    })


@pytest.fixture()
def accounts(web3):
    return web3.eth.accounts



from __future__ import absolute_import

import itertools
import logging
import time

import click

from populus.compilation import (
    compile_project_contracts,
)

from .accounts import (
    is_account_locked,
)
from .compile import (
    write_compiled_sources,
)
from .contracts import (
    verify_contract_bytecode,
)
from .observers import (
    DirWatcher,
)
from .wait import (
    Timeout,
)

from web3.utils.empty import (
    Empty,
)


def select_chain(project):
    """
    Present the user with a prompt to select which of the project chains they
    want to use.
    """
    chain_options = set(project.config['chains'].keys())

    choose_chain_msg = "\n".join(itertools.chain((
        "Available Chains",
        "----------------",
    ), (
        "{0} - {1}".format(chain_index, chain_name)
        for chain_index, chain_name in enumerate(sorted(chain_options))
    ), (
        "",
        "Enter ether the name, or number of the desired chain"
    )))
    chain_name = click.prompt(choose_chain_msg)
    if chain_name in chain_options:
        return chain_name
    elif chain_name.isdigit() and int(chain_name) < len(chain_options):
        return sorted(chain_options)[int(chain_name)]
    else:
        raise click.ClickException(
            "Invalid choice: {0}.  Please choose from one of the "
            "provided options.".format(chain_name)
        )


def select_account(chain):
    """
    Present the user with a prompt to select which of the chain accounts they
    would like to use.
    """
    all_accounts = chain.web3.eth.accounts
    if not all_accounts:
        raise click.ClickException("No accounts found on chain.")

    pick_account_message = '\n'.join(itertools.chain((
        "Accounts",
        "-----------------",
    ), (
        "{index} - {account}".format(
            account=account,
            index=index,
        ) for (index, account) in enumerate(all_accounts)
    ), (
        "",
        "Enter the account address or the number of the desired account",
    )))

    account_choice = click.prompt(
        pick_account_message,
        default=chain.web3.eth.defaultAccount or chain.web3.eth.coinbase,
    )

    if account_choice in set(all_accounts):
        return account_choice
    elif account_choice.isdigit() and int(account_choice) < len(all_accounts):
        return all_accounts[int(account_choice)]
    else:
        raise click.ClickException(
            "Invalid choice: {0}.  Please choose from one of the "
            "provided options.".format(account_choice)
        )


def request_account_unlock(chain, account, timeout):
    """
    Present a password prompt to unlock the given account.
    """
    if not is_account_locked(chain.web3, account):
        raise click.ClickException(
            "The account `{0}` is already unlocked".format(account)
        )

    unlock_account_msg = (
        "Please provide the password to unlock account `{0}`.".format(account)
    )

    # default="" is for allowing empty password
    unlock_successful = chain.web3.personal.unlockAccount(
        account,
        click.prompt(unlock_account_msg, hide_input=True, default=""),
        timeout,
    )
    if not unlock_successful:
        raise click.ClickException("Unable to unlock account: `{0}`".format(account))


def deploy_contract_and_verify(chain,
                               contract_name,
                               ContractFactory=None,
                               deploy_transaction=None,
                               deploy_args=None,
                               deploy_kwargs=None):
    """
    This is a *loose* wrapper around `populus.utils.deploy.deploy_contract`
    that handles the various concerns and logging that need to be present when
    doing this as a CLI interaction.

    Deploy a contract, displaying information about the deploy process as it
    happens.  This also verifies that the deployed contract's bytecode matches
    the expected value.
    """
    web3 = chain.web3
    logger = logging.getLogger('populus.utils.cli.deploy_contract_and_verify')

    if isinstance(web3.eth.defaultAccount, Empty):
        # must have coinbase
        # TODO: add --account arg (index or address), and set default account here if provided
        web3.eth.defaultAccount = web3.eth.coinbase

    if is_account_locked(web3, web3.eth.defaultAccount):
        try:
            chain.wait.for_unlock(web3.eth.defaultAccount or web3.eth.coinbase, 5)
        except Timeout:
            default_account = select_account(chain)
            if is_account_locked(web3, default_account):
                request_account_unlock(chain, default_account, None)
            web3.eth.defaultAccount = default_account

    logger.info("Deploying {0}".format(contract_name))

    if ContractFactory is None:
        ContractFactory = chain.provider.get_contract_factory(contract_name)

    deploy_txn_hash = ContractFactory.deploy(
        transaction=deploy_transaction,
        args=deploy_args,
        kwargs=deploy_kwargs,
    )
    deploy_txn = web3.eth.getTransaction(deploy_txn_hash)

    logger.info("Deploy Transaction Sent: {0}".format(deploy_txn_hash))
    logger.info("Waiting for confirmation...")

    contract_address = chain.wait.for_contract_address(
        deploy_txn_hash,
        timeout=180,
    )
    deploy_receipt = web3.eth.getTransactionReceipt(deploy_txn_hash)

    logger.info((
        "\n"
        "Transaction Mined\n"
        "=================\n"
        "Tx Hash      : {0}\n"
        "Address      : {1}\n"
        "Gas Provided : {2}\n"
        "Gas Used     : {3}\n\n".format(
            deploy_txn_hash,
            contract_address,
            deploy_txn['gas'],
            deploy_receipt['gasUsed'],
        )
    ))

    # Verification
    deployed_bytecode = web3.eth.getCode(contract_address)

    if ContractFactory.bytecode_runtime:
        verify_contract_bytecode(web3, ContractFactory.bytecode_runtime, contract_address)
        logger.info("Verified contract bytecode @ {0}".format(contract_address))
    else:
        logger.info(
            "No runtime available.  Falling back to verifying non-empty "
            "bytecode."
        )
        if len(deployed_bytecode) <= 2:
            logger.error("Bytecode @ {0} is unexpectedly empty.".format(contract_address))
            raise click.ClickException("Error deploying contract")
        else:
            logger.info(
                "Verified bytecode @ {0} is non-empty".format(contract_address)
            )
    return ContractFactory(address=contract_address)


def watch_project_contracts(project, compiler_settings):
    logger = logging.getLogger('populus.utils.cli.watch_project_contracts')

    def callback(file_path, event_name):
        if event_name in {'modified', 'created'}:
            logger.info("============ Compiling ==============")
            logger.info("> Change detected in: %s", file_path)
            for source_dir in project.contracts_source_dirs:
                logger.info("> Loading source files from: %s", source_dir)

            contract_source_paths, compiled_sources = compile_project_contracts(project)
            write_compiled_sources(
                project.compiled_contracts_asset_path,
                compiled_sources,
            )

            logger.info("> Watching ...")

    watcher = DirWatcher(project.contracts_dir, callback)
    watcher.start()

    try:
        while True:
            time.sleep(0.1)
    except KeyboardInterrupt:
        pass


def select_project_contract(project):
    contract_names = sorted(project.compiled_contract_data.keys())
    contract_choices = [
        " {idx}: {name}".format(
            idx=str(idx).rjust(3),
            name=name,
        ) for idx, name
        in enumerate(contract_names)
    ]
    select_contract_message = (
        "Please select the desired contract:\n\n"
        "{0}\n\n".format(
            '\n'.join(contract_choices)
        )
    )
    contract_name = click.prompt(select_contract_message)
    if contract_name in project.compiled_contract_data:
        return contract_name
    elif contract_name.isdigit() and int(contract_name) < len(contract_names):
        return contract_names[int(contract_name)]
    else:
        bad_choice_message = (
            "'{0}' is not a valid choice.  Please enter either the numeric "
            "index of the desired contract or the full name of the "
            "contract.".format(
                contract_name,
            )
        )
        raise click.ClickException(bad_choice_message)



from __future__ import absolute_import

import copy
import warnings

try:
    from contextlib import ExitStack
except ImportError:
    from contextlib2 import ExitStack

from geth import (
    DevGethProcess,
    LiveGethProcess,
    TestnetGethProcess,
    LoggingMixin,
)

from web3 import (
    IPCProvider,
    HTTPProvider,
)

from populus.utils.chains import (
    get_base_blockchain_storage_dir,
)
from populus.utils.filesystem import (
    tempdir,
)
from populus.utils.geth import (
    get_geth_logfile_path,
)

from .base import (
    BaseChain,
)


class LoggedDevGethProcess(LoggingMixin, DevGethProcess):
    def __init__(self, project_dir, blockchains_dir, chain_name, overrides):
        stdout_logfile_path = get_geth_logfile_path(
            project_dir,
            chain_name,
            'stdout'
        )
        stderr_logfile_path = get_geth_logfile_path(
            project_dir,
            chain_name,
            'stderr',
        )
        super(LoggedDevGethProcess, self).__init__(
            overrides=overrides,
            chain_name=chain_name,
            base_dir=blockchains_dir,
            stdout_logfile_path=stdout_logfile_path,
            stderr_logfile_path=stderr_logfile_path,
        )


class LoggedTestnetGethProccess(LoggingMixin, TestnetGethProcess):
    def __init__(self, project_dir, geth_kwargs):
        super(LoggedTestnetGethProccess, self).__init__(
            geth_kwargs=geth_kwargs,
        )


class LoggedMainnetGethProcess(LoggingMixin, LiveGethProcess):
    def __init__(self, project_dir, geth_kwargs):
        super(LoggedMainnetGethProcess, self).__init__(
            geth_kwargs=geth_kwargs,
            stdout_logfile_path=get_geth_logfile_path(
                project_dir,
                'mainnet',
                'stdout'
            ),
            stderr_logfile_path=get_geth_logfile_path(
                project_dir,
                'mainnet',
                'stderr',
            ),
        )


class BaseGethChain(BaseChain):
    stack = None
    geth = None

    def __init__(self, *args, **kwargs):

        warn_msg = (
            "Support for this chain will be dropped in the next populus version"
            "Populus will not run the chains, and will use the better and more robust"
            "Web3.py providers directly, as ExternalChain."
            "Please configure your chains as ExternalChain."
        )
        warnings.warn(warn_msg, DeprecationWarning)
        super(BaseGethChain, self).__init__(*args, **kwargs)

    def initialize_chain(self):
        # context manager shenanigans
        self.stack = ExitStack()
        self.geth = self.get_geth_process_instance()

    def get_web3_config(self):
        base_config = super(BaseGethChain, self).get_web3_config()
        config = copy.deepcopy(base_config)
        if not config.get('provider.settings'):
            if issubclass(base_config.provider_class, IPCProvider):
                config['provider.settings.ipc_path'] = self.geth.ipc_path
            elif issubclass(base_config.provider_class, HTTPProvider):
                config['provider.settings.endpoint_uri'] = "http://127.0.0.1:{0}".format(
                    self.geth.rpc_port,
                )
            else:
                raise ValueError("Unknown provider type")
        return config

    @property
    def geth_kwargs(self):
        return self.config.get('chain.settings', {})

    def get_geth_process_instance(self):
        raise NotImplementedError("Must be implemented by subclasses")

    def __enter__(self, *args, **kwargs):
        self.stack.enter_context(self.geth)

        if self.geth.is_mining:
            self.geth.wait_for_dag(600)
        if self.geth.ipc_enabled:
            self.geth.wait_for_ipc(60)
        if self.geth.rpc_enabled:
            self.geth.wait_for_rpc(60)

        self._running = True

        return self

    def __exit__(self, *exc_info):
        self.stack.close()
        self._running = False


class LocalGethChain(BaseGethChain):
    def get_geth_process_instance(self):
        return LoggedDevGethProcess(
            project_dir=self.project.project_dir,
            blockchains_dir=self.project.base_blockchain_storage_dir,
            chain_name=self.chain_name,
            overrides=self.geth_kwargs,
        )


class TemporaryGethChain(BaseGethChain):
    def get_geth_process_instance(self):
        tmp_project_dir = self.stack.enter_context(tempdir())
        base_blockchain_storage_dir = get_base_blockchain_storage_dir(tmp_project_dir)

        return LoggedDevGethProcess(
            project_dir=self.project.project_dir,
            blockchains_dir=base_blockchain_storage_dir,
            chain_name=self.chain_name,
            overrides=self.geth_kwargs,
        )


class TestnetChain(BaseGethChain):
    def get_geth_process_instance(self):
        return LoggedTestnetGethProccess(
            project_dir=self.project.project_dir,
            geth_kwargs=self.geth_kwargs,
        )


class MainnetChain(BaseGethChain):
    def get_geth_process_instance(self):
        return LoggedMainnetGethProcess(
            project_dir=self.project.project_dir,
            geth_kwargs=self.geth_kwargs,
        )



import copy
from populus.config import (
    load_config,
    write_config,
)

from populus.config.upgrade import (
    upgrade_config,
    ConfigContext,
)

from populus.project import (
    Project,
)

from populus.config.versions import (
    LATEST_VERSION,
    LAST_NO_USER_CONFIG_VERSION,
)

import shutil


def upgrade_configs(project_dir, logger, to_version=None):
    """upgrade project and the user config file"""

    project = Project(project_dir=project_dir, create_config_file=True)

    if to_version is None:
        to_version = LATEST_VERSION

    user_config = copy.deepcopy(project.user_config)
    if int(user_config['version']) < int(LATEST_VERSION):
        user_config = upgrade_config(user_config, ConfigContext.USER)
        write_config(user_config, project.user_config_file_path)

    if project.legacy_config_path is not None:

        legacy_config = load_config(project.legacy_config_path)
        legacy_version = legacy_config['version']
        if int(legacy_version) > int(LAST_NO_USER_CONFIG_VERSION):
            raise KeyError(
                "Unkown legacy version {legacy_version} at {legacy_config}".format(
                    legacy_version=legacy_version,
                    legacy_config=legacy_config
                )
            )
        elif int(legacy_version) < int(LAST_NO_USER_CONFIG_VERSION):
            legacy_config = upgrade_config(
                legacy_config, ConfigContext.LEGACY, LAST_NO_USER_CONFIG_VERSION
            )

        shutil.move(
            project.legacy_config_path,
            project.legacy_config_path + ".orig"
        )
        write_config(legacy_config, project.config_file_path)
        project.reload_config()

    project_config = copy.deepcopy(project.project_config)
    if int(project_config['version']) < int(LATEST_VERSION):
        project_config = upgrade_config(project_config, ConfigContext.USER)
        write_config(project_config, project.config_file_path)

    project.reload_config()
    project.clean_config()



from eth_utils import (
    is_same_address,
)

from web3.providers.tester import (
    EthereumTesterProvider,
    TestRPCProvider,
)


def is_account_locked(web3, account):
    if isinstance(web3.providers[0], (EthereumTesterProvider, TestRPCProvider)):
        return not any((is_same_address(account, a) for a in web3.eth.accounts[:10]))
    try:
        web3.eth.sign(account, 'simple-test-data')
    except ValueError as err:
        return 'account is locked' in str(err)
    else:
        return False



import logging
import time

import click

from populus.api.utils import (
    new_local_chain,
)

from populus.chain import (
    BaseGethChain,
)

from populus.utils.geth import (
    reset_chain,
)

from .main import main


@main.group('chain')
@click.pass_context
def chain_cmd(ctx):
    """
    Manage and run ethereum blockchains.
    """
    pass


@chain_cmd.command('new')
@click.argument('chain_name', nargs=1, default="local")
@click.pass_context
def chain_new(ctx, chain_name):
    """
    Create the scripts required to run a local geth blockchain
    """
    project = ctx.obj['PROJECT']
    new_local_chain(project.project_dir, chain_name)


@chain_cmd.command('reset')
@click.argument('chain_name', nargs=1, default="default")
@click.option('--confirm/--no-confirm', default=True)
@click.pass_context
def chain_reset(ctx, chain_name, confirm):
    """
    Reset a chain removing all chain data and resetting it to it's genesis
    state..
    """
    project = ctx.obj['PROJECT']

    if confirm:
        confirmation_message = (
            "Are you sure you want to reset blockchain {0}: {1}".format(
                chain_name,
                project.get_blockchain_data_dir(chain_name),
            )
        )
        if not click.confirm(confirmation_message):
            raise ctx.abort()
    reset_chain(project.get_blockchain_data_dir(chain_name))


@chain_cmd.command('run')
@click.argument('chain_name', nargs=1, default="default")
@click.option('--mine/--no-mine', default=True)
@click.option(
    '--verbosity', default=5,
    help="""
    Set verbosity of the logging output. Default is 5, Range is 0-6.
    """)
@click.pass_context
def chain_run(ctx, chain_name, mine, verbosity):
    """
    Run the named chain.
    """
    logger = logging.getLogger('populus.cli.chain.run')
    project = ctx.obj['PROJECT']

    chain = project.get_chain(chain_name)

    if isinstance(chain, BaseGethChain):
        chain.geth.register_stdout_callback(logger.info)
        chain.geth.register_stderr_callback(logger.error)

    with chain:
        try:
            while True:
                time.sleep(0.2)
        except KeyboardInterrupt:
            pass



from __future__ import absolute_import

import logging

import click

from populus.api.deploy import (
    deploy,
)

from .main import main


@main.command('deploy')
@click.option(
    'chain_name',
    '--chain',
    '-c',
    help=(
        "Specifies the chain that contracts should be deployed to. The chains "
        "mainnet' and 'morden' are pre-configured to connect to the public "
        "networks.  Other values should be predefined in your populus.ini"
    ),
)
@click.option(
    'wait_for_sync',
    '--wait-for-sync/--no-wait-for-sync',
    default=False,
    help=(
        "Determines whether the deploy command should wait until the chain is "
        "fully synced before deployment"
    ),
)
@click.argument('contracts_to_deploy', nargs=-1)
@click.pass_context
def deploy_cmd(ctx, chain_name, wait_for_sync, contracts_to_deploy):
    """
    Deploys the specified contracts to a chain.
    """
    project = ctx.obj['PROJECT']
    logger = logging.getLogger('populus.cli.deploy')

    deploy(project, logger, chain_name, wait_for_sync, contracts_to_deploy)



class PopulusException(Exception):
    pass


class PopulusResourceWarning(Warning):
    # replace with ResourceWarning when migrating to python 3
    pass


class ValidationError(PopulusException):
    pass



from __future__ import absolute_import

import logging

import click


class ClickLogHandler(logging.Handler):
    def handle(self, record):
        click.echo(self.format(record))

    def handleError(self, record):
        click.echo(self.format(record), err=True)


def get_logger_with_click_handler(name, level=None):
    if level is None:
        level = logging.INFO

    logger = logging.getLogger(name)
    logger.setLevel(level)

    if not any(isinstance(h, ClickLogHandler) for h in logger.handlers):
        logger.addHandler(ClickLogHandler())

    return logger



from __future__ import absolute_import

from semantic_version import (
    Spec,
    Version,
)

from solc import (
    get_solc_version,
)

from eth_utils import (
    is_string,
)

from .base import (
    BaseCompilerBackend,
)
from .solc_combined_json import (
    SolcCombinedJSONBackend,
)
from .solc_standard_json import (
    SolcStandardJSONBackend,
)


def get_solc_backend_class_for_version(solc_version):
    if is_string(solc_version):
        solc_version = Version(solc_version)

    if solc_version in Spec('<=0.4.8'):
        return SolcCombinedJSONBackend
    elif solc_version in Spec('>=0.4.11'):
        return SolcStandardJSONBackend
    else:
        raise OSError(
            "The installed solc compiler is not supported.  Supported versions "
            "of the solc compiler are <=0.4.8 and >=0.4.11"
        )


class SolcAutoBackend(BaseCompilerBackend):
    project_source_glob = ('*.sol', )
    test_source_glob = ('Test*.sol', )

    def __init__(self, settings):
        proxy_backend_class = get_solc_backend_class_for_version(get_solc_version())
        self.proxy_backend = proxy_backend_class(settings)

    @property
    def settings(self):
        return self.proxy_backend.settings

    def get_compiled_contracts(self, *args, **kwargs):
        return self.proxy_backend.get_compiled_contracts(*args, **kwargs)



from __future__ import absolute_import

import itertools

import anyconfig

import jsonschema

import os
import json
import logging

from cytoolz.functoolz import (
    partial,
    pipe,
)
from cytoolz.dicttoolz import (
    assoc,
)

from eth_utils import (
    to_tuple,
    is_string,
)

from populus import ASSETS_DIR
from populus.exceptions import (
    ValidationError,
)

from .contracts import (
    compute_direct_dependency_graph,
    compute_recursive_contract_dependencies,
)
from .deploy import (
    compute_deploy_order,
)
from .filesystem import (
    ensure_file_exists,
)
from .json import (
    normalize_object_for_json,
)


DEFAULT_CONTRACTS_DIR = "./contracts/"


def get_contracts_source_dirs(project_dir):
    contracts_source_dirs = [
        os.path.abspath(os.path.join(project_dir, DEFAULT_CONTRACTS_DIR)),
    ]
    return contracts_source_dirs


BUILD_ASSET_DIR = "./build"


def get_build_asset_dir(project_dir):
    build_asset_dir = os.path.join(project_dir, BUILD_ASSET_DIR)
    return build_asset_dir


COMPILED_CONTRACTS_ASSET_FILENAME = './contracts.json'


def get_compiled_contracts_asset_path(build_asset_dir):
    compiled_contracts_asset_path = os.path.join(
        build_asset_dir,
        COMPILED_CONTRACTS_ASSET_FILENAME,
    )
    return compiled_contracts_asset_path


def write_compiled_sources(compiled_contracts_asset_path, compiled_sources):
    logger = logging.getLogger('populus.compilation.write_compiled_sources')
    ensure_file_exists(compiled_contracts_asset_path)

    with open(compiled_contracts_asset_path, 'w') as outfile:
        json.dump(
            normalize_object_for_json(compiled_sources),
            outfile,
            sort_keys=True,
            indent=4,
            separators=(',', ': '),
        )

    logger.info(
        "> Wrote compiled assets to: %s",
        os.path.relpath(compiled_contracts_asset_path)
    )

    return compiled_contracts_asset_path


@to_tuple
def add_direct_dependencies_to_compiled_contracts(compiled_contracts):
    for contract_data in compiled_contracts:
        direct_dependencies = set(
            ref['name']
            for ref
            in itertools.chain(contract_data['linkrefs'], contract_data['linkrefs_runtime'])
        )
        yield assoc(contract_data, key='direct_dependencies', value=direct_dependencies)


@to_tuple
def add_full_dependencies_to_compiled_contracts(compiled_contracts):
    dependency_graph = compute_direct_dependency_graph(compiled_contracts)
    deploy_order = compute_deploy_order(dependency_graph)

    for contract_data in compiled_contracts:
        full_dependencies = compute_recursive_contract_dependencies(
            contract_data['name'],
            dependency_graph,
        )
        ordered_full_dependencies = tuple(
            contract_name
            for contract_name
            in deploy_order
            if contract_name
            in full_dependencies
        )
        yield pipe(
            contract_data,
            partial(assoc, key='full_dependencies', value=full_dependencies),
            partial(assoc, key='ordered_full_dependencies', value=ordered_full_dependencies),
        )


def post_process_compiled_contracts(compiled_contracts):
    return pipe(
        compiled_contracts,
        add_direct_dependencies_to_compiled_contracts,
        add_full_dependencies_to_compiled_contracts,
    )


def load_json_if_string(value):
    if is_string(value):
        return json.loads(value)
    else:
        return value


def normalize_contract_metadata(metadata):
    if not metadata:
        return None
    elif is_string(metadata):
        return json.loads(metadata)
    else:
        raise ValueError("Unknown metadata format '{0}'".format(metadata))


V1 = 'v1'


CONTRACT_DATA_SCHEMA_PATHS = {
    V1: os.path.join(ASSETS_DIR, 'contract-data.v1.schema.json'),
}


def validate_compiled_contracts(compiled_contracts, schema_version=V1):
    validation_errors = tuple(
        (
            contract_data['name'],
            get_contract_data_validation_errors(contract_data),
        )
        for contract_data
        in compiled_contracts
    )
    if validation_errors and any(tuple(zip(*validation_errors))[1]):
        error_messages_by_contract = tuple(
            (contract_name, "\n".join((error.message for error in errors)))
            for contract_name, errors
            in validation_errors
            if errors
        )
        error_message = (
            "The following compiled contracts did not pass validation:\n{0}".format(
                "\n".join((
                    "--------{name}-----------\n{error_messages}".format(
                        name=name,
                        error_messages=error_messages
                    ) for name, error_messages
                    in error_messages_by_contract
                ))
            )
        )
        raise ValidationError(error_message)


def get_contract_data_validation_errors(contract_data, schema_version=V1):
    if schema_version not in CONTRACT_DATA_SCHEMA_PATHS:
        raise KeyError("No schema for version: {0} - Must be one of {1}".format(
            schema_version,
            ", ".join(sorted(CONTRACT_DATA_SCHEMA_PATHS.keys())),
        ))
    schema = anyconfig.load(CONTRACT_DATA_SCHEMA_PATHS[schema_version])
    validator = jsonschema.Draft4Validator(schema)
    normalized_contract_data = normalize_object_for_json(contract_data)
    errors = [error for error in validator.iter_errors(normalized_contract_data)]
    return errors



import signal

from .wait import (
    wait_for_popen,
)


def kill_proc(proc):
    try:
        if proc.poll() is None:
            proc.send_signal(signal.SIGINT)
            wait_for_popen(proc, 5)
        if proc.poll() is None:
            proc.terminate()
            wait_for_popen(proc, 2)
        if proc.poll() is None:
            proc.kill()
            wait_for_popen(proc, 1)
    except KeyboardInterrupt:
        proc.kill()



V1 = '1'
V2 = '2'
V3 = '3'
V4 = '4'
V5 = '5'
V6 = '6'
V7 = '7'

KNOWN_LEGACY_VERSIONS = {V1, V2, V3, V4, V5, V6}
KNOWN_USER_VERSIONS = {V6, V7}

LATEST_VERSION = V7
LAST_NO_USER_CONFIG_VERSION = V6
FIRST_USER_CONFIG_VERSION = V7



from eth_utils import (
    is_boolean,
    is_number,
    is_string,
)


def is_primitive_type(value):
    return any((
        value is None,
        is_boolean(value),
        is_string(value),
        is_number(value),
    ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


setup(
    name='populllus',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='2.2.2',
    description="""Ethereum Development Framework (fork for LLL)""",
    long_description_markdown_filename='README.md',
    url='https://gitlab.com/veox/populllus',
    include_package_data=True,
    py_modules=['populus'],
    setup_requires=['setuptools-markdown'],
    install_requires=[
        "anyconfig>=0.7.0",
        "click>=6.6",
        "cytoolz>=0.8.2",
        "contextlib2>=0.5.4",
        "eth-testrpc>=1.3.0",
        "eth-utils>=0.7.4,<0.8",
        "jsonschema>=2.5.1",
        "py-geth>=1.9.0",
        "py-solc>=1.2.0",
        "pylru>=1.0.9",
        "pysha3>=0.3,!=1.0,>1.0.0",
        "pytest>=2.7.2,!=3.3.0",
        "rlp>=0.6.0,<1",
        "semantic_version>=2.6.0",
        "toposort>=1.4",
        "watchdog>=0.8.3",
        "web3>=3.16.5,<4",
    ],
    license="MIT",
    zip_safe=False,
    entry_points={
        'console_scripts': ["populus=populus.cli:main"],
        'pytest11': ['ethereum=populus.plugin'],
    },
    keywords='ethereum pytest',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
